"""The ``data_processing`` submodule provides several functions for data processing and to build annotations for yolo model.

This submodule contains the following functions :

- ``bbox2yolo(bbox:ndarray, image_height:int=1080, image_width:int=1080)`` : Function to normalize the representation of the bounding box, such that
    there are in the yolo format.
- ``build_yolo_annotations_for_images(data_dir:WindowsPath, images_dir:WindowsPath, path_bboxes:str,
                                        df_bboxes:DataFrame, df_images:DataFrame, limit_data:int,
                                         img_folder_name:str, label_folder_name:str, exclude_ids:Optional[set]=None)`` : Generates the .txt files that are necessary for yolo training.
- ``find_img_ids_to_exclude(data_dir:WindowsPath)`` : Find image ids to exclude from list of images used for building the annotation files.
- ``generate_yolo_files(output_dir:WindowsPath, train_files:List[Any,type[str]], val_files:List[Any,type[str]])`` : Generates data files for yolo training: train.txt, val.txt and data.yaml.
- ``get_annotations_from_db(password:str)`` : Gets the data from the database. Requires that your IP is configured in Azure.
- ``get_annotations_from_files(input_dir:WindowsPath, bboxes_filename:str, images_filename:str)`` : Get annotations from csv files instead of the database. The files should be located in the input_dir folder.
- ``get_train_valid(list_files:List[Any,type[str]], split:float=0.85)`` : Split data into train and validation partitions.
- ``image_orientation (image:image)`` : Function which gives the images that have a specified orientation the same orientation.
- ``plot_image_and_bboxes(img:image, anns:list, ratio:float)`` : Plots the image and the bounding box(es) associated to the detected object(s).
- ``process_annotations(anns:DataFrame, ratio:float, target_h:int=1080, target_w:int=1080)`` : Processes the annotations to match the yolo format.
- ``save_annotations_to_files(output_dir:WindowsPath, df_bboxes:DataFrame, df_images:DataFrame)`` : Save the annotations in csv format.

"""

import os
from pathlib import Path, WindowsPath
import yaml
import psycopg2
from typing import Tuple, List, Optional, Union
from datetime import datetime
from tqdm import tqdm

import numpy as np
from numpy import ndarray, array
from sklearn.model_selection import train_test_split
import pandas as pd
from pandas import DataFrame
import cv2

from matplotlib import image
import matplotlib.pyplot as plt
from PIL import Image, ExifTags, ImageDraw


class_id_to_name_mapping = {
    1: "Insulating material",
    4: "Drum",
    2: "Bottle-shaped",
    3: "Can-shaped",
    5: "Other packaging",
    6: "Tire",
    7: "Fishing net / cord",
    8: "Easily namable",
    9: "Unclear",
    0: "Tarp / fragment",
    10: "Sheet",
    11: "Black Tarp / Plastic",
}


# def plot_image_and_bboxes(img:image, anns:list, ratio:float) -> None:

#     """ Plots the image and the bounding box(es) associated to the detected object(s).

#     Args:
#         img (image): Image, from the instance file
#         anns (list): Annotations linked to the specified image, from instance file
#         ratio (float): Ratio - most often defined at the (1080/height of the image)
#     """

#     _, ax = plt.subplots(1, figsize=(12, 10))
#     ax.imshow(img)

#     for ann in anns:
#         [bbox_x, bbox_y, bbox_w, bbox_h] = (ratio*np.array(ann['bbox'])).astype(int)
#         # Obtains the new coordinates of the bboxes - normalized via the ratio.
#         rect = patches.Rectangle((bbox_x, bbox_y), bbox_w, bbox_h, linewidth=2, edgecolor='r', facecolor="none")
#         ax.add_patch(rect)

#     plt.show()
#     # Prints out a 12 * 10 image with bounding box(es).


def plot_image_and_bboxes_yolo(image: image, annotation_list: ndarray) -> None:

    """Plots the image and the bounding box(es) associated to the detected object(s).

    Args:
        image (image): Image, from the instance file
        annotation_list (ndarray): Annotations linked to the specified image, from instance file
    """

    annotations = np.array(annotation_list)

    w, h = image.size

    plotted_image = ImageDraw.Draw(image)

    transformed_annotations = np.copy(annotations)
    transformed_annotations[:, [1, 3]] = annotations[:, [1, 3]] * w
    transformed_annotations[:, [2, 4]] = annotations[:, [2, 4]] * h

    transformed_annotations[:, 1] = transformed_annotations[:, 1] - (
        transformed_annotations[:, 3] / 2
    )
    transformed_annotations[:, 2] = transformed_annotations[:, 2] - (
        transformed_annotations[:, 4] / 2
    )
    transformed_annotations[:, 3] = (
        transformed_annotations[:, 1] + transformed_annotations[:, 3]
    )
    transformed_annotations[:, 4] = (
        transformed_annotations[:, 2] + transformed_annotations[:, 4]
    )

    for ann in transformed_annotations:
        obj_cls, x0, y0, x1, y1 = ann

        plotted_image.rectangle(((x0, y0), (x1, y1)), outline="#ff8300", width=5)

        plotted_image.text((x0, y0 - 10), class_id_to_name_mapping[(int(obj_cls))])

    plt.figure(figsize=(12, 10))
    plt.imshow(np.array(image))
    plt.show()


def image_orientation(image: image) -> image:

    """Function which gives the images that have a specified orientation the same orientation.
        If the image does not have an orientation, the image is not altered.

    Args:
        image (image): Image that is in the path data_directory as well as in the instance json files

    Returns:
        image (image): image with the proper orientation
    """

    for orientation in ExifTags.TAGS.keys():
        if ExifTags.TAGS[orientation] == "Orientation":
            break
    exif = image._getexif()

    if exif is not None:
        if orientation == 274:
            return image
        elif exif[orientation] == 3:
            image = image.rotate(180, expand=True)
        elif exif[orientation] == 6:
            image = image.rotate(270, expand=True)
        elif exif[orientation] == 8:
            image = image.rotate(90, expand=True)

    return image


def bbox2yolo(
    bbox: ndarray, image_height: int = 1080, image_width: int = 1080
) -> ndarray:

    """Function to normalize the representation of the bounding box, such that
    there are in the yolo format (normalized in range [0-1]).

    Args:
        bbox (ndrray): Coordinates of the bounding box : x, y, w and h coordiates
        image_height (int, optional): Height of the image. Set as default to 1080.
        image_width (int, optional): Width of the image. Set as default to 1080.

    Returns:
        bbox (ndarray): Normalized bounding box coordinates : values between 0-1.
    """

    bbox = bbox.copy().astype(float)  # instead of np.int

    bbox[:, [0, 2]] = bbox[:, [0, 2]] / image_width
    bbox[:, [1, 3]] = bbox[:, [1, 3]] / image_height

    bbox[:, [0, 1]] = bbox[:, [0, 1]] + bbox[:, [2, 3]] / 2

    return bbox


def process_annotations(
    anns: DataFrame, ratio: float, target_h: int = 1080, target_w: int = 1080
) -> Tuple[List, ndarray]:

    """Processes the annotations to match the yolo format.

    Args:
        anns (DataFrame): image annotation informations with image id_ref and coordinates of the bounding box : x, y, w and h coordiates
        ratio (float): Ratio - most often defined at the (1080/height of the image)
        target_h (int): Height of the image. Set as default to ``1080``.
        target_w (int): Width of the image. Set as default to ``1080``.

    Returns:
        labels (List): list of object classes in the current image
        bboxes (ndarray): array of bounding boxes (x and y positions with height and width) for each object in the current image
    """

    labels = anns["id_ref_trash_type_fk"].values - 1
    bboxes = anns[["location_x", "location_y", "width", "height"]].values * ratio
    bboxes = bbox2yolo(bboxes, target_h, target_w)
    return labels, bboxes


def build_yolo_annotations_for_images(
    data_dir: WindowsPath,
    images_dir: WindowsPath,
    df_bboxes: DataFrame,
    df_images: DataFrame,
    context_filters: str = None,
    quality_filters: str = None,
    limit_data: int = 0,
    exclude_ids: Optional[set] = None,
) -> Tuple[List, int, int]:

    """Generates the .txt files that are necessary for yolo training. See
    https://github.com/ultralytics/yolov5/wiki/Train-Custom-Data for data format.

    Args:
        data_dir (WindowsPath): path of the root data directory. It should contain a folder with all useful data for images and annotations.
        images_dir (WindowsPath): path of the image directory. It should contain a folder with all images.
        path_bboxes (str): path of the bounding_boxes csv file
        df_bboxes (DataFrame): DataFrame with the bounding boxes informations (location X, Y and Height, Width)
        df_images (DataFrame): DataFrame with the image informations
        context_filters (str): the list of context filters in this format : "[context1,context2,...]". For example, `"[river,nature]"`. Set as defaults to ``None``.
        quality_filters (str): the list of quality filters in this format : "[quality1,quality2,...]". For example, `"[good,medium]"`. Set as defaults to ``None``.
        limit_data (int): limit number of images used. If you want all images set ``limit_data`` to 0.
        exclude_ids (Optional[set]): Set of image id to exclude from the process. Set as default to ``None``.

    Returns:
        valid_imagenames (List): list of image names that have been processed with success
        cpos (int): number of images with success
        cneg (int): number of images with fail
    """

    valid_imagenames = []
    input_img_folder = Path(images_dir)
    data_dir = Path(data_dir)
    list_imgs = sorted(os.listdir(input_img_folder))
    used_imgs = set(df_bboxes["id_ref_images_for_labelling"].values)

    print(f"number of images in images folder: {len(list_imgs)}")
    print(f"number of images referenced in database: {len(df_images)}")
    print(f"number of images with a bbox in database: {len(used_imgs)}")

    # apply filters if given :
    if context_filters:
        context_filters = context_filters[1:-1].split(",")
        df_images = df_images[df_images["context"].isin(context_filters)]

    if quality_filters:
        quality_filters = quality_filters[1:-1].split(",")
        df_images = df_images[df_images["image_quality"].isin(quality_filters)]

    used_imgs = used_imgs & set(df_images.index)

    print(
        f"number of images after applying context and quality filters: {len(used_imgs)}"
    )

    if exclude_ids:
        used_imgs = used_imgs - exclude_ids
        print(
            f"after exclusion, number of images with a bbox in database: {len(used_imgs)}"
        )

    if not Path.exists(data_dir / "images"):
        os.mkdir(data_dir / "images")
    if not Path.exists(data_dir / "labels"):
        os.mkdir(data_dir / "labels")

    count_exists, count_missing = 0, 0

    print("Start building the annotations ...")

    for img_id in used_imgs:

        img_name = df_images.loc[img_id]["filename"]
        if Path.exists(input_img_folder / img_name):
            count_exists += 1
            if limit_data > 0 and count_exists > limit_data:
                break

            image = Image.open(input_img_folder / img_name)

            # in place rotation of the image using Exif data
            image = image_orientation(image)

            image = np.array(image)
            h, w = image.shape[:-1]
            target_h = 1080  # the target height of the image
            ratio = target_h / h  # We get the ratio of the target and the actual height
            target_w = int(ratio * w)
            image = cv2.resize(image, (target_w, target_h))
            h, w = image.shape[:-1]

            # getting annotations and converting to yolo
            anns = df_bboxes[df_bboxes["id_ref_images_for_labelling"] == img_id]
            labels, bboxes = process_annotations(anns, ratio, target_h, target_w)
            yolo_strs = [
                str(cat) + " " + " ".join(bbox.astype(str))
                for (cat, bbox) in zip(labels, bboxes)
            ]

            # writing the image and annotation
            img_file_name = data_dir / "images" / (img_id + ".jpg")
            label_file_name = data_dir / "labels" / (img_id + ".txt")
            Image.fromarray(image).save(img_file_name)
            with open(label_file_name, "w") as f:
                f.write("\n".join(yolo_strs))

            valid_imagenames.append(img_file_name.as_posix())
        else:
            count_missing += 1

        if count_exists % 500 == 0:
            print("Exists : ", count_exists)
            print("Missing : ", count_missing)

    print(f"Process finished successfully with {count_missing} missing images !")

    return valid_imagenames, count_exists, count_missing


# def build_yolo_annotations_for_images_VM(data_dir:WindowsPath, images_dir:WindowsPath, path_bboxes:str,
#                                         df_bboxes:DataFrame, df_images:DataFrame, limit_data:int,
#                                         exclude_ids:Optional[set]=None) -> Tuple[List, int, int]:

#     """ Generates the .txt files that are necessary for yolo training. See
#     https://github.com/ultralytics/yolov5/wiki/Train-Custom-Data for data format.

#     Args:
#         data_dir (WindowsPath): path of the root data directory. It should contain a folder with all useful data for images and annotations.
#         images_dir (WindowsPath): path of the image directory. It should contain a folder with all images.
#         path_bboxes (str): path of the bounding_boxes csv file
#         df_bboxes (DataFrame): DataFrame with the bounding boxes informations (location X, Y and Height, Width)
#         df_images (DataFrame): DataFrame with the image informations
#         limit_data (int): limit number of images used. If you want all images set ``limit_data`` to 0.
#         exclude_ids (Optional[set]): Set of image id to exclude from the process. Set as default to ``None``.

#     Returns:
#         valid_imagenames (List): list of image names that have been processed with success
#         cpos (int): number of images with success
#         cneg (int): number of images with fail
#     """

#     valid_imagenames = []

#     list_imgs = sorted(os.listdir(images_dir))
#     used_imgs = set(df_bboxes["id_ref_images_for_labelling"].values)

#     print(f"number of images in images folder: {len(list_imgs)}")
#     print(f"number of images referenced in database: {len(df_images)}")
#     print(f"number of images with a bbox in database: {len(used_imgs)}")

#     if exclude_ids:
#         used_imgs = used_imgs - exclude_ids
#         print(f"after exclusion, number of images with a bbox in database: {len(used_imgs)}")

#     if not Path.exists(data_dir / "images"):
#         os.mkdir(data_dir / "images")
#     if not Path.exists(data_dir / "labels"):
#         os.mkdir(data_dir / "labels")

#     # to determine which methods or annotation processing we will use
#     # (it depends if we work with the whole dataset or not)
#     fast_ann_process = False
#     if path_bboxes.split("/")[-1].split("_")[0] == "filter":
#         fast_ann_process = True

#     count_exists, count_missing = 0, 0

#     print("Start building the annotations ...")

#     for img_id in used_imgs:
#         img_name = df_images.loc[img_id]["filename"]
#         if Path.exists(images_dir / img_name):
#             count_exists += 1
#             if limit_data > 0 and count_exists > limit_data:
#                 break

#             image = Image.open(images_dir / img_name)

#             # in place rotation of the image using Exif data
#             try :
#                 image = image_orientation(image)
#             except :
#                 pass

#             image    = np.array(image)
#             h, w     = image.shape[:-1]
#             target_h = 1080 # the target height of the image
#             ratio    = target_h / h # We get the ratio of the target and the actual height
#             target_w = int(ratio*w)
#             image    = cv2.resize(image, (target_w, target_h))
#             h, w     = image.shape[:-1]

#             # getting annotations and converting to yolo

#             if fast_ann_process:

#                 bboxes = df_bboxes[df_bboxes["id_ref_images_for_labelling"] == img_id]["bboxes"]
#                 bboxes = bboxes.iloc[0]
#                 bboxes = bboxes[1:-1].replace("'","").split(", ")
#                 yolo_strs = []
#                 for bbox in bboxes:
#                     yolo_strs.append(bbox.strip("\\n"))

#             else :

#                 anns = df_bboxes[df_bboxes["id_ref_images_for_labelling"] == img_id]
#                 labels, bboxes = process_annotations(anns, ratio, target_h, target_w)
#                 yolo_strs = [str(cat) + " " + " ".join(bbox.astype(str)) for (cat, bbox) in zip(labels, bboxes)]

#             # writing the image and annotation
#             img_file_name   = data_dir / "images" / (img_id + ".jpg")
#             label_file_name = data_dir / "labels" / (img_id + ".txt")
#             Image.fromarray(image).save(img_file_name)
#             with open(label_file_name, 'w') as f:
#                 f.write('\n'.join(yolo_strs))

#             valid_imagenames.append(img_file_name.as_posix())
#         else:
#             count_missing +=1

#         if count_exists%500==0:
#                 print("Exists : ", count_exists)
#                 print("Missing : ",count_missing)

#     print(f"Process finished successfully with {count_missing} missing images !")

#     return valid_imagenames, count_exists, count_missing


def get_train_valid(
    list_files: List[str], split: float = 0.85
) -> Tuple[List[str], List[str]]:

    """Split data into train and validation partitions.

    Args:
        list_files (List[Any,type[str]]): list of image files to split into train and test partitions
        split (float, optional): train_size between 0 and 1. Set as default to 0.85.

    Returns:
        train_files (List[Any,type[str]]): list of image names for training step
        val_files (List[Any,type[str]]): list of image names for validation step
    """

    train_files, val_files = train_test_split(list_files, train_size=split)
    train_files = list(set(train_files))
    val_files = list(set(val_files))

    return train_files, val_files


def generate_yolo_files(
    output_dir: WindowsPath, train_files: List[str], val_files: List[str]
) -> None:

    """Generates data files for yolo training: train.txt, val.txt and data.yaml.

    Args:
        output_dir (WindowsPath): path of the root data directory. It should contain a folder with all useful data for images and annotations.
        train_files (List[Any,type[str]]): list of image names for training step
        val_files (List[Any,type[str]]): list of image names for validation step
    """

    with open(output_dir / "train.txt", "w") as f:
        for path in train_files:
            f.write(path + "\n")

    with open(output_dir / "val.txt", "w") as f:
        for path in val_files:
            f.write(path + "\n")

    data = dict(
        path="./../",
        train=(output_dir / "train.txt").as_posix(),
        val=(output_dir / "val.txt").as_posix(),
        nc=10,
        names=[
            "Sheet / tarp / plastic bag / fragment",
            "Insulating material",
            "Bottle-shaped",
            "Can-shaped",
            "Drum",
            "Other packaging",
            "Tire",
            "Fishing net / cord",
            "Easily namable",
            "Unclear",
        ],
    )

    with open(output_dir / "data.yaml", "w") as outfile:
        yaml.dump(data, outfile, default_flow_style=False)


def get_annotations_from_db(password: str) -> Tuple[DataFrame, DataFrame]:

    """Gets the data from the database. Requires that your IP is configured in Azure.

    Args:
        password (str): password to connect to the SQL DataBase

    Returns:
        df_bboxes (DataFrame): DataFrame with the bounding boxes informations (location X, Y and Height, Width)
        df_images (DataFrame): DataFrame with the image informations
    """

    # Update connection string information
    host = "pgdb-plastico-prod.postgres.database.azure.com"
    dbname = "plastico-prod"
    user = "surfriderrootuser@pgdb-plastico-prod"
    password = "dbc28a2c088e4942a573c#17b1e469e83"
    sslmode = "require"

    # Construct connection string
    conn_string = (
        f"host={host} user={user} dbname={dbname} password={password} sslmode={sslmode}"
    )
    conn = psycopg2.connect(conn_string)
    print("Connection established")

    # Fetch all rows from table
    cursor = conn.cursor()

    # cursor.execute('SELECT * FROM "label".bounding_boxes')
    cursor.execute('SELECT * FROM "label".bounding_boxes_with_corrections')
    raw_annotations = cursor.fetchall()

    cursor.execute('SELECT * FROM "label".images_for_labelling')
    raw_images_info = cursor.fetchall()

    # cursor.execute('SELECT * FROM "campaign".trash_type')
    # raw_category_info = cursor.fetchall()

    df_bboxes = pd.DataFrame(
        raw_annotations,
        columns=[
            "id",
            "id_creator_fk",
            "createdon",
            "id_ref_trash_type_fk",
            "id_ref_images_for_labelling",
            "location_x",
            "location_y",
            "width",
            "height",
        ],
    )

    df_images = pd.DataFrame(
        raw_images_info,
        columns=[
            "id",
            "id_creator_fk",
            "createdon",
            "filename",
            "view",
            "image_quality",
            "context",
            "container_url",
            "blob_name",
        ],
    )
    conn.close()

    return df_bboxes, df_images.set_index("id")  # , raw_category_info


def get_annotations_from_files(
    input_dir: WindowsPath, bboxes_filename: str, images_filename: str
) -> Tuple[DataFrame, DataFrame]:

    """Get annotations from csv files instead of the database. The files should be located in the input_dir folder.

    Args:
        input_dir (WindowsPath): path of the data directory. It should contain a folder with all data (images + annotations).
        bboxes_filename (str): name of the bounding boxes csv file
        images_filename (str): name of the images csv file

    Returns:
        df_bboxes (DataFrame): DataFrame with the bounding boxes informations (location X, Y and Height, Width)
        df_images (DataFrame): DataFrame with the image informations
    """

    df_bboxes = pd.read_csv(input_dir / bboxes_filename)
    df_images = pd.read_csv(input_dir / images_filename).set_index("id")

    return df_bboxes, df_images


def save_annotations_to_files(
    output_dir: WindowsPath, df_bboxes: DataFrame, df_images: DataFrame
) -> None:

    """Saves the annotations in csv format.

    Args:
        output_dir (WindowsPath): path of the root data directory. It should contain a folder with all useful data for images and annotations.
        df_bboxes (DataFrame): DataFrame with the bounding boxes informations (location X, Y and Height, Width)
        df_images (DataFrame): DataFrame with the image informations
    """

    df_bboxes.to_csv(output_dir / "bboxes.csv")
    df_images.to_csv(output_dir / "images.csv")


def find_img_ids_to_exclude(data_dir: WindowsPath) -> set:

    """Find image ids to exclude from list of images used for building the annotation files.

    Args:
        data_dir (WindowsPath): path of the root data directory. It should contain a folder with all useful data for images and annotations.

    Returns:
        ids_to_exclude (set): set of image ids to exclude
    """

    list_files = sorted(os.listdir(Path(data_dir) / "labels"))
    ids_to_exclude = {f.split(".")[0] for f in list_files}

    return ids_to_exclude


"""-------- FONCTIONS FOR UPDATING THE DATABASE ---------"""


def convert_bboxes_to_initial_locations_from_txt_labels(
    labels_folder_path: Union[str, WindowsPath],
    img_id: str,
    target_h: int,
    ratio: float,
    target_w: int,
) -> Tuple[array, array]:

    """Convert bounding boxes to initial annotation data (location_x, location_y, Width, Height) from .txt label files.

    Args:
        labels_folder_path (Union[str,WindowsPath]): the name of the labels folder or the path od this folder
        img_id (str): the id of the current image
        target_h (int): the target height of the image
        ratio (float): the ratio of the target and the actual height
        target_w (int): the target width of the image

    Returns:
        labels (array[dtype[int64]): the array of the labels presents on the image
        bboxes (array[dtype[int64]): the coordinates of the different bounding boxes of the current image
    """

    labels_folder_path = Path(labels_folder_path)

    with open(labels_folder_path / f"{img_id}.txt") as file:
        lines = file.readlines()

    labels, bboxes = [], []

    for bbox in lines:
        bbox = bbox.split(" ")
        labels.append(bbox[0])
        bboxes.append(bbox[1:])

    labels = np.array(labels).astype(int) + 1
    bboxes = np.array(bboxes).astype(float)

    # on décentre la bbox
    bboxes[:, [0, 1]] = bboxes[:, [0, 1]] - bboxes[:, [2, 3]] / 2

    # on dénormalize
    bboxes[:, [0, 2]] = bboxes[:, [0, 2]] * target_w
    bboxes[:, [1, 3]] = bboxes[:, [1, 3]] * target_h

    # dimensions d'origine
    bboxes = bboxes / ratio

    return labels, bboxes.astype(int)


"""--------- UPDATE BOUNDING BOXES DATABASE DIRECTLY ----------"""


def update_bounding_boxes_database(
    data_dir: Union[WindowsPath, str],
    images_dir: Union[WindowsPath, str],
    labels_folder_name: Union[str, WindowsPath],
    df_bboxes: DataFrame,
    df_images: DataFrame,
) -> None:

    """Update directly the Bounding Boxes DataBase from label folder.

    Args:
        data_dir (WindowsPath): path of the root data directory. It should contain a folder with all useful data for images and annotations.
        images_dir (WindowsPath): path of the image directory. It should contain a folder with all images.
        labels_folder_name (Union[str,WindowsPath]): the name of the labels folder or the path od this folder.
        df_bboxes (DataFrame): DataFrame with the bounding boxes informations (location X, Y and Height, Width)
        df_images (DataFrame): DataFrame with the image informations
    """

    input_img_folder = Path(images_dir)
    data_dir = Path(data_dir)
    list_imgs = sorted(os.listdir(input_img_folder))
    used_imgs = set(df_bboxes["id_ref_images_for_labelling"].values)
    labels_folder_path = Path(labels_folder_name)

    modified_imgs = set(os.listdir(labels_folder_path))
    modified_ids = {img_txt.split(".")[0] for img_txt in modified_imgs}

    none_modified_imgs = used_imgs - modified_ids

    print(f"number of images in images folder: {len(list_imgs)}")
    print(f"number of images referenced in database: {len(df_images)}")
    print(f"number of images used in database: {len(used_imgs)}")
    print(f"number of images to update in database: {len(modified_ids)}")
    print(
        f"number of images to keep without updates in database: {len(none_modified_imgs)}"
    )

    count_exists = 0

    print("Start updating the annotations ...")

    # Update connection string information
    host = "pgdb-plastico-prod.postgres.database.azure.com"
    dbname = "plastico-prod"
    user = "surfriderrootuser@pgdb-plastico-prod"
    password = "dbc28a2c088e4942a573c#17b1e469e83"
    sslmode = "require"

    # Construct connection string
    conn_string = (
        f"host={host} user={user} dbname={dbname} password={password} sslmode={sslmode}"
    )
    conn = psycopg2.connect(conn_string)
    print("Connection established")

    # Insert all raow from csv file
    cursor = conn.cursor()

    for img_id in tqdm(modified_ids):

        img_name = df_images.loc[img_id]["filename"]

        infos_df_bboxes = df_bboxes[df_bboxes["id_ref_images_for_labelling"] == img_id]

        nb_trashs = len(infos_df_bboxes)

        image = Image.open(input_img_folder / img_name)

        # in place rotation of the image using Exif data
        image = image_orientation(image)

        image = np.array(image)
        h, w = image.shape[:-1]
        target_h = 1080  # the target height of the image
        ratio = target_h / h  # We get the ratio of the target and the actual height
        target_w = int(ratio * w)
        image = cv2.resize(image, (target_w, target_h))
        h, w = image.shape[:-1]

        labels, bboxes = convert_bboxes_to_initial_locations_from_txt_labels(
            labels_folder_path, img_id, target_h, ratio, target_w
        )

        row_diff = nb_trashs - len(labels)

        if row_diff < 0:

            for i in range(nb_trashs):
                row = (
                    infos_df_bboxes.iloc[i]["id"],
                    int(labels[i]),
                    img_id,
                    int(bboxes[i, 0]),
                    int(bboxes[i, 1]),
                    int(bboxes[i, 2]),
                    int(bboxes[i, 3]),
                )
                row = row[1:] + (row[0],)
                cursor.execute(
                    'UPDATE "label".bounding_boxes_with_corrections SET (id_ref_trash_type_fk, id_ref_images_for_labelling, location_x, location_y, \
                                width, height) = (%s, %s, %s, %s, %s, %s) WHERE id = %s',
                    row,
                )

            # add new rows for new trashs:
            for i in range(nb_trashs, nb_trashs + row_diff):
                row = (
                    None,
                    None,
                    datetime.now(),
                    int(labels[i]),
                    img_id,
                    int(bboxes[i, 0]),
                    int(bboxes[i, 1]),
                    int(bboxes[i, 2]),
                    int(bboxes[i, 3]),
                )
                cursor.execute(
                    'INSERT INTO "label".bounding_boxes_with_corrections(id, id_creator_fk, \
                                createdon, id_ref_trash_type_fk, id_ref_images_for_labelling, location_x, location_y, \
                                width, height) VALUES (%s, %s, %s, %s, %s, %s, %s, %s, %s)',
                    row,
                )

        elif row_diff > 0:

            for i in range(len(labels)):
                row = (
                    infos_df_bboxes.iloc[i]["id"],
                    int(labels[i]),
                    img_id,
                    int(bboxes[i, 0]),
                    int(bboxes[i, 1]),
                    int(bboxes[i, 2]),
                    int(bboxes[i, 3]),
                )
                row = row[1:] + (row[0],)
                cursor.execute(
                    'UPDATE "label".bounding_boxes_with_corrections SET (id_ref_trash_type_fk, id_ref_images_for_labelling, location_x, location_y, \
                                width, height) = (%s, %s, %s, %s, %s, %s) WHERE id = %s',
                    row,
                )
        else:

            for i in range(nb_trashs):
                row = (
                    infos_df_bboxes.iloc[i]["id"],
                    int(labels[i]),
                    img_id,
                    int(bboxes[i, 0]),
                    int(bboxes[i, 1]),
                    int(bboxes[i, 2]),
                    int(bboxes[i, 3]),
                )
                row = row[1:] + (row[0],)
                cursor.execute(
                    'UPDATE "label".bounding_boxes_with_corrections SET (id_ref_trash_type_fk, id_ref_images_for_labelling, location_x, location_y, \
                                width, height) = (%s, %s, %s, %s, %s, %s) WHERE id = %s',
                    row,
                )

        count_exists += 1
        conn.commit()

    count = cursor.rowcount
    print(count, "successful update into the bounding_boxes table.")

    # Closure of the DB connection
    cursor.close()
    conn.close()
    print("The PostgreSQL connection is close")

    print(f"Process finished successfully with {count_exists} updated images !")


"""--------- UPDATE BOUNDING BOXES DATABASE FROM CSV FILE ----------"""


def build_bboxes_csv_file_for_DB(
    data_dir: Union[WindowsPath, str],
    images_dir: Union[WindowsPath, str],
    labels_folder_name: Union[str, WindowsPath],
    df_bboxes: DataFrame,
    df_images: DataFrame,
) -> Tuple[DataFrame, List]:

    """Generates the .csv file for updating the DataBase.

    Args:
        data_dir (WindowsPath): path of the root data directory. It should contain a folder with all useful data for images and annotations.
        images_dir (WindowsPath): path of the image directory. It should contain a folder with all images.
        labels_folder_name (Union[str,WindowsPath]): the name of the labels folder or the path od this folder.
        df_bboxes (DataFrame): DataFrame with the bounding boxes informations (location X, Y and Height, Width)
        df_images (DataFrame): DataFrame with the image informations

    Returns:
        new_df_bboxes (DataFrame): new bounding boxes csv file for initial DataBase including (location X, Y and Height, Width) informations.
        exceptions (List): list of image ids with annotation errors.
    """

    input_img_folder = Path(images_dir)
    data_dir = Path(data_dir)
    list_imgs = sorted(os.listdir(input_img_folder))
    used_imgs = set(df_bboxes["id_ref_images_for_labelling"].values)
    labels_folder_path = Path(labels_folder_name)

    modified_imgs = set(os.listdir(labels_folder_path))
    modified_ids = {img_txt.split(".")[0] for img_txt in modified_imgs}

    # set of context types :
    # set_contexts = set(df_images['context'])
    # filter_context = set_contexts - {'river', 'nature'}

    none_modified_imgs = used_imgs - modified_ids
    # none_modified_imgs = set(df_images[df_images["context"].isin(filter_context)].index)

    print(f"number of images in images folder: {len(list_imgs)}")
    print(f"number of images referenced in database: {len(df_images)}")
    print(f"number of images used in database: {len(used_imgs)}")
    print(f"number of images to update in database: {len(modified_ids)}")
    print(
        f"number of images to keep without updates in database: {len(none_modified_imgs)}"
    )

    count_exists = 0

    new_df_bboxes = df_bboxes[
        df_bboxes["id_ref_images_for_labelling"].isin(none_modified_imgs)
    ]

    index = 2 * len(new_df_bboxes)

    print("Start updating the annotations ...")

    for img_id in tqdm(modified_ids):

        img_name = df_images.loc[img_id]["filename"]

        infos_df_bboxes = df_bboxes[df_bboxes["id_ref_images_for_labelling"] == img_id]
        nb_trashs = len(infos_df_bboxes)

        image = Image.open(input_img_folder / img_name)

        # in place rotation of the image using Exif data

        image = image_orientation(image)

        image = np.array(image)
        h, w = image.shape[:-1]
        target_h = 1080  # the target height of the image
        ratio = target_h / h  # We get the ratio of the target and the actual height
        target_w = int(ratio * w)
        image = cv2.resize(image, (target_w, target_h))
        h, w = image.shape[:-1]

        (labels, bboxes,) = convert_bboxes_to_initial_locations_from_txt_labels(
            labels_folder_path, img_id, target_h, ratio, target_w
        )

        row_diff = nb_trashs - len(labels)

        if row_diff < 0:

            for i in range(nb_trashs):
                index = index + 1
                new_df_bboxes.loc[index] = [
                    infos_df_bboxes.iloc[i]["id"],
                    infos_df_bboxes.iloc[i]["id_creator_fk"],
                    infos_df_bboxes.iloc[i]["createdon"],
                    labels[i],
                    img_id,
                    bboxes[i, 0],
                    bboxes[i, 1],
                    bboxes[i, 2],
                    bboxes[i, 3],
                ]

            index += 20
            # add new rows for new trashs:
            for i in range(nb_trashs, nb_trashs + row_diff):
                index = index + 1
                new_df_bboxes.loc[index] = [
                    None,
                    None,
                    datetime.now(),
                    labels[i],
                    img_id,
                    bboxes[i, 0],
                    bboxes[i, 1],
                    bboxes[i, 2],
                    bboxes[i, 3],
                ]

            index += 20

        elif row_diff > 0:

            for i in range(len(labels)):
                index = index + 1
                new_df_bboxes.loc[index] = [
                    infos_df_bboxes.iloc[i]["id"],
                    infos_df_bboxes.iloc[i]["id_creator_fk"],
                    infos_df_bboxes.iloc[i]["createdon"],
                    labels[i],
                    img_id,
                    bboxes[i, 0],
                    bboxes[i, 1],
                    bboxes[i, 2],
                    bboxes[i, 3],
                ]
            index += 20
        else:

            for i in range(nb_trashs):
                index = index + 1
                new_df_bboxes.loc[index] = [
                    infos_df_bboxes.iloc[i]["id"],
                    infos_df_bboxes.iloc[i]["id_creator_fk"],
                    infos_df_bboxes.iloc[i]["createdon"],
                    labels[i],
                    img_id,
                    bboxes[i, 0],
                    bboxes[i, 1],
                    bboxes[i, 2],
                    bboxes[i, 3],
                ]
            index += 20

        count_exists += 1

    print(f"Process finished successfully with {count_exists} updated images !")

    return new_df_bboxes


def fill_bounding_boxes_table_with_corrections(
    new_csv_bounding_boxes: Union[WindowsPath, str]
) -> None:

    """Fill the bounding boxes DataBase from scratch. Requires that your IP is configured in Azure.

    Args:
        new_csv_bounding_boxes (Union[WindowsPath,str]) : the path of the bounding boxes csv files with annotation corrections.
    """

    # Update connection string information
    host = "pgdb-plastico-prod.postgres.database.azure.com"
    dbname = "plastico-prod"
    user = "surfriderrootuser@pgdb-plastico-prod"
    password = "dbc28a2c088e4942a573c#17b1e469e83"
    sslmode = "require"

    # Construct connection string
    conn_string = (
        f"host={host} user={user} dbname={dbname} password={password} sslmode={sslmode}"
    )
    conn = psycopg2.connect(conn_string)
    print("Connection established")

    # Insert all raow from csv file
    cursor = conn.cursor()

    new_df_bboxes = pd.read_csv(new_csv_bounding_boxes)

    for i in tqdm(range(len(new_df_bboxes))):

        row = tuple(new_df_bboxes.loc[i])
        row = tuple([int(val) if type(val) != str else val for val in row])

        cursor.execute(
            'INSERT INTO "label".bounding_boxes_with_corrections(id, id_creator_fk, \
            createdon, id_ref_trash_type_fk, id_ref_images_for_labelling, location_x, location_y, \
                width, height) VALUES (%s, %s, %s, %s, %s, %s, %s, %s, %s)',
            row,
        )

        conn.commit()
        count = cursor.rowcount
    print(count, "successful inserts into the bounding_boxes table.")

    # Closure of the DB connection
    cursor.close()
    conn.close()
    print("The PostgreSQL connection is close")


def update_bounding_boxes_table_with_corrections(
    new_csv_bounding_boxes,
) -> None:

    """Update the bounding boxes DataBase from a csv file with annotation corrections. Requires that your IP is configured in Azure.

    Args:
        new_csv_bounding_boxes (Union[WindowsPath,str]) : the path of the bounding boxes csv files with annotation corrections.
    """

    # Update connection string information
    host = "pgdb-plastico-prod.postgres.database.azure.com"
    dbname = "plastico-prod"
    user = "surfriderrootuser@pgdb-plastico-prod"
    password = "dbc28a2c088e4942a573c#17b1e469e83"
    sslmode = "require"

    # Construct connection string
    conn_string = (
        f"host={host} user={user} dbname={dbname} password={password} sslmode={sslmode}"
    )
    conn = psycopg2.connect(conn_string)
    print("Connection established")

    # Insert all raow from csv file
    cursor = conn.cursor()

    new_df_bboxes = pd.read_csv(new_csv_bounding_boxes)

    for i in tqdm(range(len(new_df_bboxes))):

        row = tuple(new_df_bboxes.loc[i])
        row = tuple(
            [int(val) if (type(val) != str and val is not None) else val for val in row]
        )

        if row[0] is None:

            cursor.execute(
                'INSERT INTO "label".bounding_boxes_with_corrections(id, id_creator_fk, \
                createdon, id_ref_trash_type_fk, id_ref_images_for_labelling, location_x, location_y, \
                    width, height) VALUES (%s, %s, %s, %s, %s, %s, %s, %s, %s)',
                row,
            )

        else:

            row = row[1:] + (row[0],)
            cursor.execute(
                'UPDATE "label".bounding_boxes_with_corrections SET (id_creator_fk, \
                createdon, id_ref_trash_type_fk, id_ref_images_for_labelling, location_x, location_y, \
                    width, height) = (%s, %s, %s, %s, %s, %s, %s, %s) WHERE id = %s',
                row,
            )

        conn.commit()
    count = cursor.rowcount
    print(count, "successful update into the bounding_boxes table.")

    # Closure of the DB connection
    cursor.close()
    conn.close()
    print("The PostgreSQL connection is close")



"""The ``trackers`` submodule provides several functions to build trackers for videos.

This submodule contains the following classes:
- ``Tracker`` : Build a Tracker for videos.
- ``EKF(Tracker)`` : Extended Kalman Filter based on Tracker class.

This submodule contains the following functions:
- ``get_tracker(algorithm_and_params:str)`` : Provide specific tracker according algorithm and parameters.

"""

import matplotlib.patches as mpatches
import numpy as np
from numpy import array, ndarray, dtype, float64
from typing import Any, Tuple
from cv2 import Mat
from pykalman import KalmanFilter
from scipy.stats import multivariate_normal

from plasticorigins.tracking.utils import in_frame


class Tracker:

    """Build a Tracker for videos.

    Args:
        frame_nb (int): the number of the frame
        X0 (array[Any,dtype[float64]]): Array of state mean
        confidence (float): confidence in [0,1]
        class_id (int): id of the object class
        transition_variance (ndarray[Any,dtype[float64]]): Array of the transition variance
        observation_variance (ndarray[Any,dtype[float64]]): Array of the observation variance
        delta (float): the probability threshold of an object to belong to a specific neighbourhood
    """

    def __init__(
        self,
        frame_nb: int,
        X0: array,
        confidence: float,
        class_id: int,
        transition_variance: ndarray[Any, dtype[float64]],
        observation_variance: ndarray[Any, dtype[float64]],
        delta: float,
    ):
        self.transition_covariance = np.diag(transition_variance)
        self.observation_covariance = np.diag(observation_variance)
        self.updated = False
        self.steps_since_last_observation = 0
        self.enabled = True
        self.tracklet = [(frame_nb, X0, confidence, class_id)]
        self.delta = delta

    def store_observation(
        self,
        observation: array,
        frame_nb: int,
        confidence: float,
        class_id: int,
    ) -> None:

        """Method to store observations from frames.

        Args:
            observation (array): observation of the current frame
            frame_nb (int): numero of the current frame
            confidence (float): confidence of an observation
            class_id (int): object class id
        """

        self.tracklet.append((frame_nb, observation, confidence, class_id))
        self.updated = True

    def update_status(self, flow: Mat) -> None:

        """Update status from input flow.

        Args:
            flow (Mat): the input flow for updating status
        """

        if self.enabled and not self.updated:
            self.steps_since_last_observation += 1
            self.enabled = self.update(None, None, None, flow)

        else:
            self.steps_since_last_observation = 0

        self.updated = False

    def build_confidence_function(self, flow: Mat) -> Mat:

        """Build confidence function from flow.

        Args:
            flow (array): the input flow

        Returns:
            The computing confidence distribution based on predictive distribution from flow.
        """

        def confidence_from_multivariate_distribution(
            coord: array, distribution: array
        ) -> Mat:

            """Computes confidence from multivariate distribution.

            Args:
                coord (array): coordinates of the local object
                distribution (array): mathematical input distribution

            Returns:
                The computing confidence distribution.
            """

            delta = self.delta
            x = coord[0]
            y = coord[1]
            right_top = np.array([x + delta, y + delta])
            left_low = np.array([x - delta, y - delta])
            right_low = np.array([x + delta, y - delta])
            left_top = np.array([x - delta, y + delta])

            return (
                distribution.cdf(right_top)
                - distribution.cdf(right_low)
                - distribution.cdf(left_top)
                + distribution.cdf(left_low)
            )

        distribution = self.predictive_distribution(flow)

        return lambda coord: confidence_from_multivariate_distribution(
            coord, distribution
        )

    def cls_score_function(self, conf: float, label: int) -> float:

        """Generates a score based on classes associated with observation in this tracker.

        Args:
            conf (float): confidence of the label
            label (int): label id

        Returns:
            A score based on classes associated with observation.
        """

        class_conf = sum(tr[2] for tr in self.tracklet if tr[3] == label)
        other_conf = sum(tr[2] for tr in self.tracklet)

        return (class_conf + conf) / (other_conf + conf)

    def get_display_colors(self, display: Any, tracker_nb: int) -> array:

        """Get the display colors for a tracker.

        Args:
            display (Any): display from tracker
            tracker_nb (int): the number of the current tracker

        Returns:
            colors (array) : The colors for the current tracker
        """

        colors = display.colors
        color = colors[tracker_nb % len(colors)]
        display.legends.append(mpatches.Patch(color=color, label=len(self.tracklet)))

        return colors[tracker_nb % len(colors)]


class EKF(Tracker):

    """Extended Kalman Filter : infinite impulse response filter that estimates the states of a dynamic system
    from a series of incomplete or noisy measurements.

    Args:
        frame_nb (int): the number of the frame
        X0 (array[Any,dtype[float64]]): Array of initial state mean
        confidence (float): confidence in [0,1]
        class_id (int): id of the object class
        transition_variance (ndarray[Any,dtype[float64]]): Array of the transition variance
        observation_variance (ndarray[Any,dtype[float64]]): Array of the observation variance
        delta (float): the probability threshold of an object to belong to a specific neighbourhood
    """

    def __init__(
        self,
        frame_nb: int,
        X0: array,
        confidence: float,
        class_id: int,
        transition_variance: ndarray[Any, dtype[float64]],
        observation_variance: ndarray[Any, dtype[float64]],
        delta: float,
    ):
        super().__init__(
            frame_nb,
            X0,
            confidence,
            class_id,
            transition_variance,
            observation_variance,
            delta,
        )
        self.filter = KalmanFilter(
            initial_state_mean=X0,
            initial_state_covariance=self.observation_covariance,
            transition_covariance=self.transition_covariance,
            observation_matrices=np.eye(2),
            observation_covariance=self.observation_covariance,
        )

        self.filtered_state_mean = X0
        self.filtered_state_covariance = self.observation_covariance

    def get_update_parameters(self, flow: Mat) -> Tuple[array, array]:

        """Update parameters from input flow.

        Args:
            flow (Mat): the input flow for updating status

        Returns:
            The updated flow value with gradient operation.
        """

        flow_value = flow[
            int(self.filtered_state_mean[1]),
            int(self.filtered_state_mean[0]),
            :,
        ]

        grad_flow_value = np.array(
            [np.gradient(flow[:, :, 0]), np.gradient(flow[:, :, 1])]
        )[
            :,
            :,
            int(self.filtered_state_mean[1]),
            int(self.filtered_state_mean[0]),
        ]
        return (
            np.eye(2) + grad_flow_value,
            flow_value - grad_flow_value.dot(self.filtered_state_mean),
        )

    def EKF_step(self, observation: array, flow: Mat) -> tuple:

        """Perform a one-step update to estimate the state at time ``t+1`` given an observation
        at time ``t+1`` and the previous estimate for time ``t`` given observations from times ``[0...t]``.
        This method is useful if one wants to track an object with streaming observations.

        Args:
            observation (array): the input observation for EKF at time ``t+1``
            flow (Mat): the input flow for EKF

        Returns:
            EKF_step (tuple) : Perform a one-step update to estimate the state at time ``t+1``
        """

        transition_matrix, transition_offset = self.get_update_parameters(flow)

        return self.filter.filter_update(
            self.filtered_state_mean,
            self.filtered_state_covariance,
            transition_matrix=transition_matrix,
            transition_offset=transition_offset,
            observation=observation,
        )

    def update(
        self,
        observation: array,
        confidence: float,
        class_id: int,
        flow: Mat,
        frame_nb: int = None,
    ) -> bool:

        """Enable the update according the EKF results from the current step.

        Args:
            observation (array): current observation of the frame
            confidence (float): confidence of the observation
            class_id (int): the object class in the observation
            flow (Mat): the input flow for updating status
            frame_nb (int): the numero of the frame. Set as default to ``None``

        Returns:
            enabled (bool): Enable the update if ``True``, not if ``False``
        """

        if observation is not None:
            self.store_observation(observation, frame_nb, confidence, class_id)

        (
            self.filtered_state_mean,
            self.filtered_state_covariance,
        ) = self.EKF_step(observation, flow)

        enabled = (
            False if not in_frame(self.filtered_state_mean, flow.shape[:-1]) else True
        )

        return enabled

    def predictive_distribution(self, flow: Mat) -> array:

        """Build predictive distribution from flow.

        Args:
            flow (Mat): the input flow

        Returns:
            The computed predictive distribution from normal multivariate model.
        """

        filtered_state_mean, filtered_state_covariance = self.EKF_step(None, flow)

        distribution = multivariate_normal(
            filtered_state_mean,
            filtered_state_covariance + self.observation_covariance,
            allow_singular=True,
        )

        return distribution

    def fill_display(self, display: Any, tracker_nb: int) -> None:

        """Fill display of a specific tracker.

        Args:
            display (Any): the display for tracker
            tracker_nb (int): the number of tracker
        """

        yy, xx = np.mgrid[
            0 : display.display_shape[1] : 1, 0 : display.display_shape[0] : 1
        ]
        pos = np.dstack((xx, yy))
        distribution = multivariate_normal(
            self.filtered_state_mean, self.filtered_state_covariance
        )

        color = self.get_display_colors(display, tracker_nb)
        cs = display.ax.contour(distribution.pdf(pos), colors=color)
        display.ax.clabel(cs, inline=True, fontsize="large")
        display.ax.scatter(
            self.filtered_state_mean[0],
            self.filtered_state_mean[1],
            color=color,
            marker="x",
            s=100,
        )


trackers = {"EKF": EKF}


def get_tracker(algorithm_and_params: str) -> Tracker:

    """Provide specific tracker according algorithm and parameters.

    Args:
        algorithm_and_params (str): specify the algorithm and the parameters used for tracker. You may specify only the algorithm.

    Returns:
        tracker (Tracker): Set the specific tracker according algorithm and parameters
    """

    print(f"{algorithm_and_params} will be used for tracking.")

    splitted_name = algorithm_and_params.split("_")
    if len(splitted_name) > 1:
        algorithm_name, param = splitted_name
        tracker = trackers[algorithm_name]
        tracker.set_param(param)

    else:
        algorithm_name = splitted_name[0]
        tracker = trackers[algorithm_name]

    return tracker



"""The ``yolo`` submodule provides utils functions for YOLO model.

This submodule contains the following classes:

- ``DetectTorchScript`` : Torch script yolo class.

This submodule contains the following functions:

- ``load_model(model_path:str, device:str, conf:float=0.35, iou:float=0.50)`` : Load yolo model from model_path.
- ``predict_yolo(model, img:Mat, size:int=768, augment:bool=False)`` : Interpret yolo prediction object.
- ``voc2centerdims(bboxes:ndarray)`` : Compute center coordinates of the bounding boxes and find their size (width, height).

"""

import json
import cv2
from cv2 import Mat
import numpy as np
from numpy import ndarray, dtype, int64, float64
import torch
import torch.nn as nn
import yaml
import yolov5
from yolov5.models.common import Detections
from yolov5.utils.general import non_max_suppression
from typing import Any, Tuple, Union, Dict, Optional

# This has to be kept for now as this depends on the model training
id_categories = {
    1: "Insulating material",
    4: "Drum",
    2: "Bottle-shaped",
    3: "Can-shaped",
    5: "Other packaging",
    6: "Tire",
    7: "Fishing net / cord",
    8: "Easily namable",
    9: "Unclear",
    0: "Sheet / tarp / plastic bag / fragment",
}

categories_id = {v: k for k, v in id_categories.items()}
get_id = lambda cat: categories_id[cat]


def load_model(
    model_path: str, device: str, conf: float = 0.35, iou: float = 0.50
) -> Any:

    """Load yolo model from model_path.

    Args:
        model_path (str): the model path
        device (str): type of device used for computing ("cpu", "cuda",...)
        conf (float): confidence value in [0,1]. Set as default to 0.35.
        iou (float): iou value in [0,1]. Set as default to 0.50.

    Returns:
        model (Any): The loaded model with right input parameters
    """

    model = yolov5.load(model_path, device=device)
    model.conf = conf
    model.iou = iou
    model.classes = None
    model.multi_label = False
    model.max_det = 1000

    return model


def voc2centerdims(bboxes: ndarray) -> ndarray:

    """Compute center coordinates of the bounding boxes and find their size (width, height).

    Args:
        bboxes (ndarray): contain all positions of the bounding boxes : ``voc -> [x1, y1, x2, y2]``

    Returns:
        bboxes (ndarray): the updates bboxes with the center coordinates and (width, height) : ``output -> [x_center, y_center, w, h]``
    """

    bboxes[..., 2:4] -= bboxes[..., 0:2]  # find w,h
    bboxes[..., 0:2] += bboxes[..., 2:4] / 2  # find center

    return bboxes


def predict_yolo(
    model, img: Mat, size: int = 768, augment: bool = False
) -> Tuple[
    ndarray[Any, dtype[int64]],
    ndarray[Any, dtype[float64]],
    ndarray[Any, dtype[int64]],
]:

    """Interpret yolo prediction object.

    Args:
        model (Any): yolo model used for prediction
        img (Mat): image to predict
        size (int): image size. Set as default to 768.
        augment (bool): Data augmentation if True. Set as default to False.

    Returns:
       bboxes (ndarray[Any,dtype[int64]]): The bounding boxes with center coordinates and W, H. Empty array if bboxes is empty.
       confs (ndarray[Any,dtype[float64]]): The confidence values from predictions. Empty array if bboxes is empty.
       labels (ndarray[Any,dtype[int64]]): The list of label ids from predictions. Empty array if bboxes is empty.
    """

    img = cv2.cvtColor(img, cv2.COLOR_BGR2RGB)
    results = model(img, size=size, augment=augment)
    preds = results.pandas().xyxy[0]
    bboxes = preds[["xmin", "ymin", "xmax", "ymax"]].values  # voc format

    if len(bboxes):
        bboxes = voc2centerdims(bboxes)
        bboxes = bboxes.astype(int)
        confs = preds.confidence.values
        labels = np.array(list(map(get_id, preds.name.values)))
        return bboxes, confs, labels

    else:
        return np.array([]), np.array([]), np.array([])


class DetectTorchScript(nn.Module):

    """Torch script yolo class.

    Args:
        weights (ndarray): weights of the object detection model
        conf (float): confidence value. Set as default to ``0.35``.
        iou (float): Set as default to ``0.5``.
        class_names (Optional[Dict]): dictionnary of the object classes for detection in this format : `{class_name : class_id,...}`. Set as default to ``None``.
        data (Optional[str]): path of the yaml file data. Set as default to ``None``.
    """

    max_det = 1000
    agnostic_nms = True
    classes = None

    # YOLOv5 TorchScript class for python inference
    def __init__(
        self,
        weights: ndarray,
        conf: float = 0.35,
        iou: float = 0.50,
        class_names: Optional[Dict] = None,
        data: Optional[str] = None,
    ):
        super().__init__()
        self.class_names = class_names
        self.categories_id = {v: k for k, v in class_names.items()}

        self.conf_thres = conf
        self.iou_thres = iou

        self.size = 640
        stride, names = (
            64,
            [f"class{i}" for i in range(1000)],
        )  # assign defaults

        if data:  # data.yaml path (optional)
            with open(data, errors="ignore") as f:
                names = yaml.safe_load(f)["names"]  # class names

        extra_files = {"config.txt": ""}  # model metadata
        model = torch.jit.load(weights, _extra_files=extra_files)

        if extra_files["config.txt"]:
            d = json.loads(extra_files["config.txt"])  # extra_files dict
            stride, names = int(d["stride"]), d["names"]
        self.__dict__.update(locals())  # assign all variables to self

    def forward(self, image: ndarray) -> Union[torch.Tensor, nn.Module]:

        """Evaluate the yolo model predictions for the input image.

        Args:
            image (ndarray): expects image as numpy array of shape (B x H x W x C) or (H x W x C). In practice B=1 and H x W = 640.

        Returns:
            The model predictions.
        """

        shape = image.shape

        if len(shape) == 3:
            image = np.expand_dims(image, 0)  # add batch axis
        # x = cv2.cvtColor(image, cv2.COLOR_BGR2RGB)
        # image = letterbox(image, new_shape=self.size, auto=False)[0] # pad and reshape

        x = np.ascontiguousarray(
            image[:, :, :, ::-1].transpose((0, 3, 1, 2))
        )  # BHWC to BCHW, and BGR to RGB
        x = torch.from_numpy(x) / 255  # uint8 to fp16/32

        return self.model(x)[0]

    def get_label_idxs(self, preds_names: ndarray) -> ndarray:

        """Give the class ids knowing the class names.

        Args:
            preds_names (ndarray): list of predicted class names

        Returns:
            The array of the labels ids found in the predictions.
        """

        get_id = lambda cat: self.categories_id[cat]

        return np.array(list(map(get_id, preds_names)))

    def detect(
        self, images: ndarray
    ) -> Tuple[
        ndarray[Any, dtype[int64]],
        ndarray[Any, dtype[float64]],
        ndarray[Any, dtype[int64]],
    ]:

        """Interpret yolo prediction object with batch size set to 1.

        Args:
            images (ndarray): images to predict

        Returns:
            bboxes (ndarray[Any,dtype[int64]]): The bounding boxes with center coordinates and W, H. Empty array if bboxes is empty.
            confs (ndarray[Any,dtype[float64]]): The confidence values from predictions. Empty array if bboxes is empty.
            labels (ndarray[Any,dtype[int64]]): The list of label ids from predictions. Empty array if bboxes is empty.
        """

        pred = self.forward(images)
        pred2 = non_max_suppression(
            pred,
            conf_thres=self.conf_thres,
            iou_thres=self.iou_thres,
            classes=self.classes,
            agnostic=self.agnostic_nms,
            max_det=self.max_det,
        )
        detections = Detections(
            images,
            pred2,
            None,
            times=(0, 0, 0, 0),
            names=self.class_names,
            shape=(1, 3, 640, 640),
        )
        preds = detections.pandas().xyxy[0]
        bboxes = preds[["xmin", "ymin", "xmax", "ymax"]].values  # voc format

        if len(bboxes):
            bboxes = voc2centerdims(bboxes)
            bboxes = bboxes.astype(int)
            confs = preds.confidence.values
            labels = self.get_label_idxs(preds.name.values)
            return bboxes, confs, labels

        else:
            return np.array([]), np.array([]), np.array([])



"""The ``coco_utils`` submodule provides several classes and functions to convert coco objects.

This submodule allows the user to :

- filter and remap coco categories
- convert coco objects (polygons) to masks (torch tensors)
- convert coco objects (polygons) to bounding boxes
- check validated conversion

This submodule contains the following classes:

- ``CocoDetectionWithExif`` : Perform Coco detection with exif.
- ``ConvertCocoPolysToBboxes`` : Convert Coco Polygons to bounding boxes.
- ``ConvertCocoPolysToMask`` : Convert Coco Polygons to Masks.
- ``FilterAndRemapCocoCategories`` : Filter and remap Coco categories.
- ``GroupInTensor`` : Group arrays in (torch) Tensors.

This submodule contains the following functions:

- ``coco_remove_images_without_annotations(dataset:Dataset, cat_list:Optional[List[Dict]]=None)`` : Check if the annotation are valid or not empty.
- ``convert_coco_poly_to_mask(segmentations:List[Any], height:int, width:int)`` : Converting Coco polygons to torch masks (tensors).
- ``get_coco(root:str, image_set:str, transforms:Callable)`` : Transform images to a dataset with specified transformation.
- ``get_surfrider(root:str, image_set:str, transforms:Callable)`` : Transform images to a dataset with specified transformation.
- ``get_surfrider_video_frames(root:str, image_set:str, transforms:Callable)`` : Transform images to a dataset with specified transformation.

"""

import copy
import os
from typing import Any, Callable, Optional, Tuple, List, Dict
from numpy import ndarray
import imageio
import torch
from torch.utils.data import Dataset
import torchvision
from PIL import Image
from pycocotools import mask as coco_mask

from .transforms import Compose


class FilterAndRemapCocoCategories:

    """Filter and remap Coco categories.

    Args:
        categories (List): list of object classes
        remap (bool): Remapping if ``True``, ``False`` if not
    """

    def __init__(self, categories: List, remap: bool = True):
        self.categories = categories
        self.remap = remap

    def __call__(self, image: Image, anno: List[Dict]) -> Tuple[Any, List[Dict]]:

        """Remapping the images with true category ids.

        Args:
            image (Image): Image, from the instance file
            anno (List[Dict]): Annotations linked to the specified image, from instance file

        Returns:
            image (Image): Image, from the instance file
            anno (List[Dict]): Remapped annotations linked to the specified image, from instance file
        """

        anno = [obj for obj in anno if obj["category_id"] in self.categories]
        if not self.remap:
            return image, anno
        anno = copy.deepcopy(anno)
        for obj in anno:
            obj["category_id"] = self.categories.index(obj["category_id"])
        return image, anno


def convert_coco_poly_to_mask(
    segmentations: List[Any], height: int, width: int
) -> torch.Tensor:

    """Converting Coco polygons to masks (torch tensors).

    Args:
        segmentations (List): list of polygons
        height (int): height of the mask
        width (int): width of the mask

    Returns:
        masks (torch.Tensor): output masks
    """

    masks = []

    for polygons in segmentations:
        rles = coco_mask.frPyObjects(polygons, height, width)
        mask = coco_mask.decode(rles)

        if len(mask.shape) < 3:
            mask = mask[..., None]

        mask = torch.as_tensor(mask, dtype=torch.uint8)
        mask = mask.any(dim=2)
        masks.append(mask)

    if masks:
        masks = torch.stack(masks, dim=0)

    else:
        masks = torch.zeros((0, height, width), dtype=torch.uint8)

    return masks


class ConvertCocoPolysToMask:

    """Convert Coco Polygons to Masks."""

    def __call__(self, image: Image, anno: List[Dict]) -> Tuple[Any, Any]:

        """Generating target image from masks.

        Args:
            image (Image): Image, from the instance file
            anno (List[Dict]): Annotations linked to the specified image, from instance file

        Returns:
            image (Image): Image, from the instance file
            target (Image): Target image resulting from the product between categories and masks
        """

        w, h = image.size
        segmentations = [obj["segmentation"] for obj in anno]
        cats = [obj["category_id"] for obj in anno]

        if segmentations:
            masks = convert_coco_poly_to_mask(segmentations, h, w)
            cats = torch.as_tensor(cats, dtype=masks.dtype)
            # merge all instance masks into a single segmentation map
            # with its corresponding categories
            target, _ = (masks * cats[:, None, None]).max(dim=0)
            # discard overlapping instances
            target[masks.sum(0) > 1] = 255

        else:
            target = torch.zeros((h, w), dtype=torch.uint8)

        target = Image.fromarray(target.numpy())

        return image, target


class ConvertCocoPolysToBboxes:

    """Convert Coco Polygons to bounding boxes."""

    def __call__(self, image: Image, anno: List[Dict]) -> Tuple[Any, Dict]:

        """Building dictionnary of image annotations.

        Args:
            image (Image): Image, from the instance file
            anno (List[Dict]): Annotations linked to the specified image, from instance file

        Returns:
            image (Image): Image, from the instance file
            target (Dict): Target dictionnary with bounding boxes associated with object classes
        """

        bboxes = [obj["bbox"] for obj in anno]
        cats = [obj["category_id"] for obj in anno]

        if bboxes:
            target = {"bboxes": bboxes, "cats": cats}

        else:
            target = {"bboxes": [], "cats": []}

        return image, target


class GroupInTensor:

    """Group arrays in (torch) Tensors."""

    def __call__(
        self, frames: ndarray, ground_truths: ndarray
    ) -> Tuple[torch.Tensor, torch.Tensor]:

        """Convert multi-dimensionnal numpy array frames and ground truths into torch tensors.

        Args:
            frames (ndarray): list of frames
            ground_truths (ndarray): the multi-dimensionnal ground truths array

        Returns:
            Both ``frames`` and ``ground_truths`` convert into torch tensors.
        """

        return torch.from_numpy(frames), torch.from_numpy(ground_truths)


def coco_remove_images_without_annotations(
    dataset: Dataset, cat_list: Optional[List[Dict]] = None
) -> Dataset:

    """Remove images without annotations.

    Args:
        dataset (Dataset): The image dataset to process.
        cat_list (Optional[List[Dict]]): The list of object categories. Set as default to ``None``.

    Returns:
        The filtered dataset.
    """

    def has_valid_annotation(anno: List[Dict]) -> bool:

        """Check if the annotations are valid or not empty.

        Args:
            anno (List[Dict]): Annotations linked to the specified image, from instance file

        Returns:
            ``True`` if the annotations exist and are valid.
            Otherwise, ``False``.
        """

        # if it's empty, there is no annotation
        if len(anno) == 0:
            return False

        # if more than 1k pixels occupied in the image
        return sum(obj["area"] for obj in anno) > 1000

    assert isinstance(dataset, torchvision.datasets.CocoDetection)

    ids = []

    for ds_idx, img_id in enumerate(dataset.ids):
        ann_ids = dataset.coco.getAnnIds(imgIds=img_id, iscrowd=None)
        anno = dataset.coco.loadAnns(ann_ids)

        if cat_list:
            anno = [obj for obj in anno if obj["category_id"] in cat_list]

        if has_valid_annotation(anno):
            ids.append(ds_idx)

    dataset = torch.utils.data.Subset(dataset, ids)

    return dataset


def get_coco(root: str, image_set: str, transforms: Callable) -> Dataset:

    """Transform images to a dataset with specified transformation (classes ``FilterAndRemapCocoCategories``
        ``ConvertCocoPolysToMask`` and ``_coco_remove_images_without_annotations`` used).

    Args:
        root (str): Root directory where images are downloaded to
        image_set (str): directory to annotation data. Two possible values : ``"train"`` or ``"val"``.
        transforms (Callable): A function/transformation that takes input sample and its target as entry and returns a transformed version

    Returns:
        dataset (Dataset): dataset of images with specific transformations
    """

    PATHS = {
        "train": (
            "train2017",
            os.path.join("annotations", "instances_train2017.json"),
        ),
        "val": (
            "val2017",
            os.path.join("annotations", "instances_val2017.json"),
        ),
        # "train": ("val2017", os.path.join("annotations", "instances_val2017.json"))
    }
    CAT_LIST = [
        0,
        5,
        2,
        16,
        9,
        44,
        6,
        3,
        17,
        62,
        21,
        67,
        18,
        19,
        4,
        1,
        64,
        20,
        63,
        7,
        72,
    ]

    transforms = Compose(
        [
            FilterAndRemapCocoCategories(CAT_LIST, remap=True),
            ConvertCocoPolysToMask(),
            transforms,
        ]
    )

    img_folder, ann_file = PATHS[image_set]
    img_folder = os.path.join(root, img_folder)
    ann_file = os.path.join(root, ann_file)

    dataset = torchvision.datasets.CocoDetection(
        img_folder, ann_file, transforms=transforms
    )

    if image_set == "train":
        dataset = coco_remove_images_without_annotations(dataset, CAT_LIST)

    return dataset


def get_surfrider(root: str, image_set: str, transforms: Callable) -> Dataset:

    """Transform images to a dataset with specified transformation (class ``CocoDetectionWithExif`` used).

    Args:
        root (str): Root directory where images are downloaded to
        image_set (str): directory to annotation data. Two possible values : ``"train"`` or ``"val"``.
        transforms (Callable): A function/transform that takes input sample and its target as entry and returns a transformed version.

    Returns:
        dataset (Dataset): dataset of images with specific transformations.
    """

    PATHS = {
        "train": (
            "Images_md5",
            os.path.join("annotations", "instances_train.json"),
        ),
        "val": (
            "Images_md5",
            os.path.join("annotations", "instances_val.json"),
        ),
        # "train": ("val2017", os.path.join("annotations", "instances_val2017.json"))
    }

    transforms = Compose(
        [
            # FilterAndRemapCocoCategories(CAT_LIST, remap=True),
            ConvertCocoPolysToMask(),
            # ConvertCocoPolysToBboxes(),
            transforms,
        ]
    )

    img_folder, ann_file = PATHS[image_set]
    img_folder = os.path.join(root, img_folder)
    ann_file = os.path.join(root, ann_file)

    dataset = torchvision.datasets.CocoDetection(
        img_folder, ann_file, transforms=transforms
    )

    # if image_set == "train":
    #     dataset = _coco_remove_images_without_annotations(dataset, CAT_LIST)

    return dataset


def get_surfrider_video_frames(
    root: str, image_set: str, transforms: Callable
) -> Dataset:

    """Transform images to a dataset with specified transformation (class `CocoDetection used`).

    Args:
        root (str): Root directory where images are downloaded to
        image_set (str): directory to annotation data. Two possible values : ``"train"`` or ``"val"``.
        transforms (Callable): A function/transform that takes input sample and its target as entry and returns a transformed version.

    Returns:
        dataset (Dataset): dataset of images with specific transformations
    """

    PATHS = {
        "train": (
            "images",
            os.path.join("annotations", "instances_train.json"),
        ),
        "val": ("images", os.path.join("annotations", "instances_val.json")),
        # "train": ("val2017", os.path.join("annotations", "instances_val2017.json"))
    }
    # CAT_LIST = [0, 1, 2, 3]

    transforms = Compose(
        [
            # FilterAndRemapCocoCategories(CAT_LIST, remap=True),
            # ConvertCocoPolysToMask(),
            ConvertCocoPolysToBboxes(),
            transforms,
        ]
    )

    img_folder, ann_file = PATHS[image_set]
    img_folder = os.path.join(root, img_folder)
    ann_file = os.path.join(root, ann_file)

    dataset = CocoDetectionWithExif(img_folder, ann_file, transforms=transforms)

    # if image_set == "train":
    #     dataset = _coco_remove_images_without_annotations(dataset, CAT_LIST)

    return dataset


class CocoDetectionWithExif(torchvision.datasets.CocoDetection):

    """Perform Coco detection with exif.

    Args:
        root (str): path of root folder / data folder
        annFile (str): name of annotation file
        transform (Optional[Callable]): transformation for detection. Set as default to ``None``.
        target_transform (Optional[Callable]): target transformation for detection. Set as default to ``None``.
        transforms (Optional[Callable]): transformations for detection. Set as default to ``None``.
    """

    def __init__(
        self,
        root: str,
        annFile: str,
        transform: Optional[Callable] = None,
        target_transform: Optional[Callable] = None,
        transforms: Optional[Callable] = None,
    ):
        super().__init__(root, annFile, transform, target_transform, transforms)

    def __getitem__(self, index: int) -> Tuple[ndarray, ndarray]:

        """Apply a transformation to the image which corresponds to the given index.

        Args:
            index (int): Index of item

        Returns:
            Tuple ``(image, target)`` with ``target`` is the object returned by ``coco.loadAnns``
        """

        coco = self.coco
        img_id = self.ids[index]
        ann_ids = coco.getAnnIds(imgIds=img_id)
        target = coco.loadAnns(ann_ids)

        path = coco.loadImgs(img_id)[0]["file_name"]
        img = imageio.imread(os.path.join(self.root, path))

        if self.transforms is not None:
            img, target = self.transforms(img, target)

        return img, target



"""The ``transforms`` submodule provides several transformation classes to convert, process and normalize images.

This submodule contains the following classes:

- ``Compose`` : Compose class for creating transformation pipeline on images.
- ``Normalize`` : Apply normalization on images with specified mean and standard deviation.
- ``ToTensorBboxes`` : Transform tensors to bounding boxes.
- ``TrainTransforms`` : Apply transformations on images for training.
- ``TransformFrames`` : Transform frames (with normalization).
- ``ValTransforms`` : Apply transformations on images for validation.

"""

import cv2
import imgaug as ia
import numpy as np
from numpy import ndarray
from PIL import Image
import torch
import torchvision.transforms as T
from imgaug import augmenters as iaa
from imgaug.augmentables.bbs import BoundingBox, BoundingBoxesOnImage
from torchvision.transforms import functional as F
from typing import Any, Callable, Tuple, List, Dict, Union

from plasticorigins.tools.misc import blob_for_bbox

ia.seed(1)


class Compose:

    """Compose class for creating transformation pipeline on images.

    Args:
        transforms (Callable): transformations to apply
    """

    def __init__(self, transforms: Callable):
        self.transforms = transforms

    def __call__(self, image: Image, target: Any) -> Tuple[torch.Tensor, torch.Tensor]:

        """Apply transformations on image according to the given target.

        Args:
            image (Image): Image, from the instance file
            target (Any): target for transformations

        Returns:
            Tuple (image, target) after transformations
        """

        for t in self.transforms:
            image, target = t(image, target)

        return image, target


class ToTensorBboxes:

    """Transform tensors to bounding boxes.

    Args:
        num_classes (int): number of different object classes
        downsampling_factor (float): downsampling factor for rescaling images
    """

    def __init__(self, num_classes: int, downsampling_factor: float):
        self.num_classes = num_classes
        self.downsampling_factor = downsampling_factor

    def __call__(
        self, image: Union[Any, ndarray], bboxes: List[Dict]
    ) -> Tuple[torch.Tensor, torch.Tensor]:

        """Apply transformations on image according to the given target.

        Args:
            image (Union[Image,ndarray]): PIL Image, from the instance file or ndarray.
            bboxes (List[Dict]): List of annotations with labels and bounding boxes coordinates

        Returns:
            Tuple (image, target) after transformations
        """

        h, w = image.shape[:-1]
        image = F.to_tensor(image)

        if self.downsampling_factor is not None:
            blobs = np.zeros(
                shape=(
                    self.num_classes + 2,
                    h // self.downsampling_factor,
                    w // self.downsampling_factor,
                )
            )
        else:
            blobs = np.zeros(shape=(self.num_classes + 2, h, w))

        for bbox_imgaug in bboxes:
            cat = bbox_imgaug.label - 1
            bbox = [
                bbox_imgaug.x1,
                bbox_imgaug.y1,
                bbox_imgaug.width,
                bbox_imgaug.height,
            ]

            new_blobs, ct_int = blob_for_bbox(
                bbox, blobs[cat], self.downsampling_factor
            )

            blobs[cat] = new_blobs

            if ct_int is not None:
                ct_x, ct_y = ct_int

                if ct_x < blobs.shape[2] and ct_y < blobs.shape[1]:
                    blobs[-2, ct_y, ct_x] = bbox[3]
                    blobs[-1, ct_y, ct_x] = bbox[2]

        target = torch.from_numpy(blobs)

        return image, target


class Normalize:

    """Apply normalization on images with specified mean and standard deviation.

    Args:
        mean (List[float]): mean of normalization function
        std (List[float]): standard deviation of normalization function
    """

    def __init__(self, mean: List[float], std: List[float]):
        self.mean = mean
        self.std = std

    def __call__(
        self, image: torch.Tensor, target: Any
    ) -> Tuple[torch.Tensor, torch.Tensor]:

        """Normalize the input image.

        Args:
            image (torch.Tensor): Image, from the instance file in torch Tensor format
            target (Any): target for normalization

        Returns:
            Tuple (image, target) after normalization
        """

        image = F.normalize(image, mean=self.mean, std=self.std)

        return image, target


class TrainTransforms:

    """Apply transformations on images for training.

    Args:
        base_size (int): base size of images
        crop_size (Tuple[int,int]): crop size of images ``(H, W)``
        num_classes (int): number of object classes
        downsampling_factor (float): downsampling factor
        hflip_prob (float): High flip probability. Set as default to ``0.5``
        mean (Tuple[float,float,float]) : mean of normalization function. Set as default to ``(0.485, 0.456, 0.406)``
        std (Tuple[float,float,float]) : standard deviation of normalization function. Set as default to ``(0.229, 0.224, 0.225)``
    """

    def __init__(
        self,
        base_size: int,
        crop_size: Tuple[int, int],
        num_classes: int,
        downsampling_factor: float,
        hflip_prob: float = 0.5,
        mean: Tuple[float, float, float] = (0.485, 0.456, 0.406),
        std: Tuple[float, float, float] = (0.229, 0.224, 0.225),
    ):
        self.num_classes = num_classes
        self.downsampling_factor = downsampling_factor
        self.base_size = base_size
        self.crop_height, self.crop_width = crop_size
        self.hflip_prob = hflip_prob
        self.random_size_range = (
            int(self.base_size),
            int(2.0 * self.base_size),
        )
        self.seq = iaa.Sequential(
            [
                iaa.Resize(
                    {
                        "height": self.random_size_range,
                        "width": "keep-aspect-ratio",
                    }
                ),
                iaa.Fliplr(p=self.hflip_prob),
                iaa.PadToFixedSize(width=self.crop_width, height=self.crop_height),
                iaa.CropToFixedSize(width=self.crop_width, height=self.crop_height),
            ]
        )
        self.last_transforms = Compose(
            [
                ToTensorBboxes(num_classes, downsampling_factor),
                Normalize(mean=mean, std=std),
            ]
        )

    def __call__(self, image: ndarray, target: Any) -> Any:

        """Create a new augmenter sequence and apply transformations based on the argument ``last_transforms``.

        Args:
            image (ndarray): Image, from the instance file
            target (Any): target for transformations

        Returns:
            The last transformations required.
        """

        bboxes_imgaug = [
            BoundingBox(
                x1=bbox[0],
                y1=bbox[1],
                x2=bbox[0] + bbox[2],
                y2=bbox[1] + bbox[3],
                label=cat,
            )
            for bbox, cat in zip(target["bboxes"], target["cats"])
        ]
        bboxes = BoundingBoxesOnImage(bboxes_imgaug, shape=image.shape)

        image, bboxes_imgaug = self.seq(image=image, bounding_boxes=bboxes)

        return self.last_transforms(image, bboxes_imgaug)


class ValTransforms:

    """Apply transformations on images for validation.

    Args:
        base_size (Tuple[int,int]): base size of images ``(H, W)``
        crop_size (Tuple[int,int]): crop size of images ``(H, W)``
        num_classes (int): number of object classes
        downsampling_factor (float): downsampling factor
        mean (Tuple[float,float,float]) : mean of normalization function. Set as default to ``(0.485, 0.456, 0.406)``
        std (Tuple[float,float,float]) : standard deviation of normalization function. Set as default to ``(0.229, 0.224, 0.225)``
    """

    def __init__(
        self,
        base_size: Tuple[int, int],
        crop_size: Tuple[int, int],
        num_classes: int,
        downsampling_factor: float,
        mean: Tuple[float, float, float] = (0.485, 0.456, 0.406),
        std: Tuple[float, float, float] = (0.229, 0.224, 0.225),
    ):
        self.num_classes = num_classes
        self.downsampling_factor = downsampling_factor
        self.base_size = base_size
        self.crop_height, self.crop_width = crop_size
        self.seq = iaa.Sequential(
            [
                iaa.Resize(
                    {
                        "height": int(self.base_size),
                        "width": "keep-aspect-ratio",
                    }
                ),
                # iaa.Rotate((-45,45)),
                iaa.CenterPadToFixedSize(
                    width=self.crop_width, height=self.crop_height
                ),
                iaa.CenterCropToFixedSize(
                    width=self.crop_width, height=self.crop_height
                ),
            ]
        )
        self.last_transforms = Compose(
            [
                ToTensorBboxes(num_classes, downsampling_factor),
                Normalize(mean=mean, std=std),
            ]
        )

    def __call__(self, image: ndarray, target: Dict) -> Any:

        """Create a new augmenter sequence and apply transformations based on the argument last_transforms.

        Args:
            image (ndarray): Image, from the instance file
            target (Dict): target for transformations (ground truths ``gt`` or heat map ``hm``)

        Returns:
            The last transformations required.
        """

        bboxes_imgaug = [
            BoundingBox(
                x1=bbox[0],
                y1=bbox[1],
                x2=bbox[0] + bbox[2],
                y2=bbox[1] + bbox[3],
                label=cat,
            )
            for bbox, cat in zip(target["bboxes"], target["cats"])
        ]
        bboxes = BoundingBoxesOnImage(bboxes_imgaug, shape=image.shape)

        image, bboxes_imgaug = self.seq(image=image, bounding_boxes=bboxes)

        return self.last_transforms(image, bboxes_imgaug)


class TransformFrames:

    """Transform frames (with normalization)."""

    def __init__(self):
        transforms = []

        transforms.append(T.Lambda(lambda img: cv2.cvtColor(img, cv2.COLOR_BGR2RGB)))
        transforms.append(T.ToTensor())
        transforms.append(
            T.Normalize(mean=[0.485, 0.456, 0.406], std=[0.229, 0.224, 0.225])
        )

        self.transforms = T.Compose(transforms)

    def __call__(self, image: ndarray) -> Any:

        """Apply given transformations on the input image.

        Args:
            image (ndarray): Image, from the instance file

        Returns:
            The image with the transformations required.
        """

        return self.transforms(image)



"""The ``optical_flow`` submodule provides several functions for computing dense optical flow.

This submodule contains the following functions:

- ``compute_flow(frame0:Mat, frame1:Mat, downsampling_factor:Union[float,int])`` : Resizes the input frames and computes a dense optical flow using ``flow_opencv_dense``.
- ``flow_opencv_dense(img:Mat, img2:Mat)`` : Computes a dense optical flow using the Gunnar Farneback's algorithm.

"""

import cv2
from cv2 import Mat
from typing import Union


def flow_opencv_dense(img: Mat, img2: Mat) -> Mat:

    """Computes a dense optical flow using the Gunnar Farneback's algorithm.

    Args:
        img (Mat): first input image as a numpy matrix
        img2 (Mat): second input image of the same size and the same type as img, as a numpy matrix

    Returns:
        flow (Mat): the dense optical flow for both images
    """

    prvs = cv2.cvtColor(img, cv2.COLOR_BGR2GRAY)
    next = cv2.cvtColor(img2, cv2.COLOR_BGR2GRAY)

    flow = cv2.calcOpticalFlowFarneback(prvs, next, None, 0.5, 3, 15, 3, 5, 1.2, 0)

    return flow


def compute_flow(
    frame0: Mat, frame1: Mat, downsampling_factor: Union[float, int]
) -> Mat:

    """Resizes the input frames and computes a dense optical flow using ``flow_opencv_dense``.

    Args:
        frame0 (Mat): first input frame as a matrix
        frame1 (Mat): second input frame of the same size and the same type as ``frame0``
        downsampling_factor (Union[float,int]) : downsampling factor to resize the frames

    Returns:
        The dense optical flow for both frames.
    """

    if downsampling_factor > 1:
        h, w = frame0.shape[:-1]
        new_h = h // downsampling_factor
        new_w = w // downsampling_factor

        frame0 = cv2.resize(frame0, (new_w, new_h))
        frame1 = cv2.resize(frame1, (new_w, new_h))

    return flow_opencv_dense(frame0, frame1)



"""The ``video_readers`` submodule provides several classes for reading videos and frames.

This submodule contains the following classes:

- ``AdvancedFrameReader`` : Advanced reader for frames with specific parameters.
- ``IterableFrameReader`` : Iterable reader for frames.
- ``SimpleVideoReader`` : Simple reader for frames.
- ``TorchIterableFromReader`` : Torch Iterable reader for frames.

"""

import cv2
from cv2 import Mat
import torch
from tqdm import tqdm
from numpy import ndarray
from typing import Callable, List, Tuple, Union, Optional


def square_crop(input_frame: Mat, out_shape: Tuple[int, int]):

    """Crops the largest square in the center of the image.

    Args:
        input_frame (Mat): input image in ``cv2 Mat`` format.
        out_shape (Tuple[int, int]): the output shape of the resized image.

    Returns:
        The resized image with ``out_shape`` shape.
    """

    h, w, _ = input_frame.shape

    if h > w:
        new_h = w
        xtop = h // 2 - new_h // 2
        crop = input_frame[xtop : xtop + new_h, :, :]

    elif h < w:
        new_w = h
        yleft = w // 2 - new_w // 2
        crop = input_frame[:, yleft : yleft + new_w, :]

    else:
        crop = input_frame

    return cv2.resize(crop, out_shape)


class AdvancedFrameReader:

    """Advanced reader for frames with specific parameters.

    Args:
        video_name (str): name of the video / path to the video
        read_every (int): parameter set to ``1`` for reading the whole video (all the frames). ``0`` means "to skip".
        rescale_factor (Union[int,float]): rescale factor for frames
        init_time_min (int): number of minutes of the initial time
        init_time_s (int): number of seconds of the initial time
    """

    def __init__(
        self,
        video_name: str,
        read_every: int,
        rescale_factor: Union[int, float],
        init_time_min: int,
        init_time_s: int,
    ):
        self.cap = cv2.VideoCapture(video_name)

        self.original_width = self.cap.get(cv2.CAP_PROP_FRAME_WIDTH)
        self.original_height = self.cap.get(cv2.CAP_PROP_FRAME_HEIGHT)

        if rescale_factor != 1:
            self.set_rescale_factor(rescale_factor)
        else:
            self.original_shape_mode = True

        self.init_rescale_factor = rescale_factor

        self.frame_skip = read_every - 1
        self.fps = self.cap.get(cv2.CAP_PROP_FPS) / read_every
        print(f"Reading at {self.fps:.2f} fps")

        self.set_time_position(init_time_min, init_time_s)
        self.init_frame = self.cap.get(cv2.CAP_PROP_POS_FRAMES)
        self.total_num_frames = self.cap.get(cv2.CAP_PROP_FRAME_COUNT)

    def post_process(self, ret: bool, frame: Mat) -> Tuple[bool, Mat]:

        """Apply a post resize process on the read image if necessary (according the ret value).

        Args:
            ret (bool): if the ``read()`` function reads successfully the capture, ``ret = True``. Else, ``ret = False``.
            frame (Mat): frame to post process (resize)

        Returns:
            ret (bool): the same `ret` value as the input
            resized_frame (Mat): the resized frame or the original frame. If `ret` is ``False``, returns empty matrix.
        """

        if ret:
            if self.original_shape_mode:
                return ret, frame
            else:
                return ret, cv2.resize(frame, self.new_shape)
        else:
            return ret, []

    def skip(self) -> None:

        """Skip if there are no frames to read."""

        if not self.nb_frames_read:
            return
        else:
            for _ in range(self.frame_skip):
                self.cap.read()

    def read_frame(self) -> Tuple[bool, Mat]:

        """Read a frame.

        Returns:
            ret (bool): First output of the ``post_process`` method : if the ``read()`` function reads successfully the frame, ``ret = True``. Else, ``ret = False``.
            resized_frame (Mat): Second output of the ``post_process`` method : the resized frame or the original frame. If `ret` is False, returns empty matrix.
        """

        self.skip()
        ret, frame = self.cap.read()
        self.nb_frames_read += 1

        return self.post_process(ret, frame)

    def set_time_position(self, time_min: int, time_s: int) -> None:

        """Set time positions in seconds.

        Args:
            time_min (int): the number of minutes for the time position
            time_s (int): the number of seconds for the time position
        """

        time = 60 * time_min + time_s
        self.cap.set(cv2.CAP_PROP_POS_MSEC, 1000 * time)
        self.init_frame = self.cap.get(cv2.CAP_PROP_POS_FRAMES)
        # print('Reading from {}min{}sec'.format(time_min,time_s))
        self.nb_frames_read = 0

    def reset_init_frame(self) -> None:

        """Reset to the initial frame."""

        self.cap.set(cv2.CAP_PROP_POS_FRAMES, int(self.init_frame))
        self.nb_frames_read = 0

    def set_init_frame(self, init_frame: Union[int, str]) -> None:

        """Set and start at the initial frame.

        Args:
            init_frame (Union[int,str]): the number of the initial frame
        """

        self.cap.set(cv2.CAP_PROP_POS_FRAMES, int(init_frame))
        self.init_frame = self.cap.get(cv2.CAP_PROP_POS_FRAMES)
        self.nb_frames_read = 0

    def set_rescale_factor(self, rescale_factor: Union[int, float]) -> None:

        """Rescale the size of a frame (width, height).

        Args:
            rescale_factor (Union[int,float]): the rescale factor
        """

        width = int(self.original_width / rescale_factor)
        height = int(self.original_height / rescale_factor)
        self.new_shape = (width, height)
        self.original_shape_mode = False
        # print('Reading in {}x{}'.format(width, height))

    def set_original_shape_mode(self, mode: bool) -> None:

        """Set the original shape mode.

        Args:
            mode (bool): the original shape mode is kept if ``mode = True``
        """

        self.original_shape_mode = mode

    def reset_init_rescale_factor(self) -> None:

        """Reset the initial rescale factor."""

        self.set_rescale_factor(self.init_rescale_factor)


class IterableFrameReader:

    """Iterable reader for frames.

    Args:
        video_filename (str): name of the video / path to the video
        skip_frames (int): parameter for skipping frames. Set as default to``0``. It means "read every".
        output_shape (Optional[Tuple[int,int]]): the shape of the outputs (frames). Set as default to ``None``.
        progress_bar (bool): to display the progress bar. Set as default to ``False``.
        preload (bool): If we want to preload all the frames. Set as default to ``False``,
        max_frame (int): the maximum number of frames to read. Set as default to ``0``.
    """

    def __init__(
        self,
        video_filename: str,
        skip_frames: int = 0,
        output_shape: Optional[Tuple[int, int]] = None,
        progress_bar: bool = False,
        preload: bool = False,
        max_frame: int = 0,
        crop: Optional[ndarray] = None,
    ):
        # store arguments for reset
        self.video_filename = video_filename
        self.max_frame_arg = max_frame
        self.progress_bar_arg = progress_bar
        self.preload = preload
        self.skip_frames = skip_frames
        self.crop = crop

        self.video = cv2.VideoCapture(video_filename)
        self.input_shape = (
            int(self.video.get(cv2.CAP_PROP_FRAME_WIDTH)),
            int(self.video.get(cv2.CAP_PROP_FRAME_HEIGHT)),
        )
        self.total_num_frames = int(self.video.get(cv2.CAP_PROP_FRAME_COUNT))

        self.max_num_frames = (
            min(max_frame, self.total_num_frames)
            if max_frame != 0
            else self.total_num_frames
        )
        self.num_skipped_frames = 0
        self.counter = 0
        self.progress_bar = None

        if output_shape is None:
            w, h = self.input_shape
            new_h = (h | 31) + 1
            new_w = (w | 31) + 1
            self.output_shape = (new_w, new_h)
        else:
            self.output_shape = output_shape

        self.fps = self.video.get(cv2.CAP_PROP_FPS) / (self.skip_frames + 1)

        if self.preload:
            self.frames = self._load_all_frames()

    def update_progress_bar(self) -> None:

        """Update the bar progression about the the reading of the frames."""

        if self.progress_bar_arg:

            if self.progress_bar:
                # update_progress_bar
                self.progress_bar.update()

            else:
                # create progress bar
                self.progress_bar = tqdm(
                    total=int(self.max_num_frames / (self.skip_frames + 1)),
                    position=1,
                    leave=True,
                )

    def reset_video(self) -> None:

        """Reset the video. This method is needed as ``cv2.CAP_PROP_POS_FRAMES``
        does not work on all backends.
        """

        self.video.release()
        if self.progress_bar:
            self.progress_bar.close()
        self.__init__(
            self.video_filename,
            self.skip_frames,
            self.output_shape,
            self.progress_bar_arg,
            self.preload,
            self.max_frame_arg,
            self.crop,
        )

    def _load_all_frames(self) -> List[Mat]:

        """Load all frames after their reading.

        Returns:
            frames (List[Mat]): list of the frames
        """

        frames = []

        while True:
            ret, frame = self._read_frame()
            if ret:
                frames.append(frame)
            else:
                break

        if self.progress_bar:
            self.progress_bar.reset()

        return frames

    def __next__(self) -> Union[Mat, StopIteration]:

        """Attempt of reading a frame.

        Returns:
            The read frame or a stop iteration status
        """

        self.counter += 1

        if self.preload:
            if self.counter < len(self.frames):
                frame = self.frames[self.counter]
                self.update_progress_bar()
                return frame

        else:
            if self.counter < self.max_num_frames:
                ret, frame = self._read_frame()
                if ret:
                    return frame
                else:
                    self.num_skipped_frames += 1
                    return next(self)

        self.reset_video()

        raise StopIteration

    def _read_frame(self) -> Tuple[bool, Mat]:

        """Read a frame.

        Returns:
            ret (bool): If the ``read()`` function reads successfully the video, ``ret = True``. Else, ``ret = False``.
            frame (Mat): The resized frame or the original frame if not skipped
        """

        ret, frame = self.video.read()
        self._skip_frames()

        if ret:
            self.update_progress_bar()

            if self.crop:
                frame = square_crop(frame, self.output_shape)

            else:
                frame = cv2.resize(frame, self.output_shape)

        return ret, frame

    def __iter__(self) -> None:
        return self

    def _skip_frames(self) -> None:

        """Skip frames."""

        for _ in range(self.skip_frames):
            self.counter += 1
            self.video.read()

    def get_inv_mapping(self, downsampling_factor: float):

        """Returns a mapping between coordinates in cropped space
        and coordinates in the original video"""

        h_in, w_in = self.input_shape
        x_top, y_left = 0, 0

        if self.crop:
            if h_in > w_in:
                h_new, w_new = w_in, w_in
                x_top = h_in // 2 - h_new // 2
                y_left = 0
            elif h_in < w_in:
                h_new, w_new = h_in, h_in
                y_left = w_in // 2 - w_new // 2
                x_top = 0
            else:
                h_new, w_new = w_in, w_in
        else:
            h_new, w_new = h_in, w_in

        ratio_x, ratio_y = (
            h_new * downsampling_factor / self.output_shape[0],
            w_new * downsampling_factor / self.output_shape[1],
        )

        mapping = lambda x, y: (int(x * ratio_x + x_top), int(y * ratio_y + y_left))
        return mapping


class SimpleVideoReader:

    """Simple reader for frames.

    Args:
        video_filename (str): name of the video / path to the video
        skip_frames (int): parameter for skipping frames. Set as default to``0``. It means "read every".
    """

    def __init__(self, video_filename: str, skip_frames: int = 0):
        self.skip_frames = skip_frames
        self.video = cv2.VideoCapture(video_filename)
        self.shape = (
            int(self.video.get(cv2.CAP_PROP_FRAME_WIDTH)),
            int(self.video.get(cv2.CAP_PROP_FRAME_HEIGHT)),
        )
        self.fps = self.video.get(cv2.CAP_PROP_FPS) / (skip_frames + 1)
        self.frame_nb = 0
        self.num_frames = self.video.get(cv2.CAP_PROP_FRAME_COUNT)

    def read(self) -> Tuple[bool, Mat, int]:

        """Reads a frame and increments the `frame_nb' of read frames.

        Returns:
            ret (bool): If the ``read()`` function reads successfully the video, ``ret = True``. Else, ``ret = False``.
            frame (Mat): the read frame
            frame_nb (int): the number of already read frames
        """

        ret, frame = self.video.read()
        self.frame_nb += 1
        self._skip_frames()

        return ret, frame, self.frame_nb - 1

    def set_frame(self, frame_nb_to_set: int) -> None:

        """Set ``frame_nb_to_set`` frames.

        Args:
            frame_nb_to_set (int): the number of frames to set
        """

        self.video.set(cv2.CAP_PROP_POS_FRAMES, frame_nb_to_set)
        self.frame_nb = frame_nb_to_set

    def _skip_frames(self) -> None:

        """Skip frames."""

        for _ in range(self.skip_frames):
            self.video.read()


class TorchIterableFromReader(torch.utils.data.IterableDataset):

    """Torch Iterable reader for frames.

    Args:
        reader (torch.utils.data.IterableDataset): the dataset of read images
        transforms (Callable): the specific transformations for frames
    """

    def __init__(self, reader: torch.utils.data.IterableDataset, transforms: Callable):
        self.transforms = transforms
        self.reader = reader

    def __iter__(self) -> None:

        """Iterable transformations on frames."""

        for frame in self.reader:
            if frame is not None:
                yield self.transforms(frame)



import pandas as pd
import matplotlib.pyplot as plt


def annotations_cat(df_bboxes):

    """Function returning a plot and table with the proportion of trash by category.

    Args:
        df_bboxes(csv file): Csv file with the bounding boxes description, coordinates, ID and linked image.

    Returns:
        plot(plot): Plot of the count of the categories x-axis : label category, y-axis count.
        new_table(dataframe): ID, percentage and count of the different trash categories.

    """

    categories = [
        "N/A",
        "Sheet / tarp / plastic bag / fragment",
        "Insulating material",
        "Drum",
        "Bottle-shaped",
        "Can-shaped",
        "Other packaging",
        "Tire",
        "Fishing net / cord",
        "Easily namable",
        "Unclear",
    ]
    # trash categories

    cat_percentages = round(
        (df_bboxes["id_ref_trash_type_fk"].value_counts() / len(df_bboxes) * 100),
        2,
    ).to_frame()
    # percentages of each trash category in the dataset

    table = pd.DataFrame(categories)  # creation of data frame of the categories
    table["Percentage"] = cat_percentages  # adding the percentages to the df
    table["ID"] = list(range(0, 11))  # adding the ID to the df
    table["Count"] = (
        df_bboxes["id_ref_trash_type_fk"].value_counts().to_frame()
    )  # adding the count of the labels to the df
    table.rename(columns={0: "Category"}, inplace=True)  # rename the category column
    new_table = table.set_index("ID").drop([0])  # final df, dropping excess info

    fig, ax = plt.subplots(figsize=(15, 7))  # creation
    dfg = df_bboxes.groupby(["id_ref_trash_type_fk"]).size()
    plot = dfg.plot(
        kind="bar",
        title="Trash Size by ID",
        ylabel="Size",
        xlabel="Trash ID",
        figsize=(6, 5),
    )
    ax.legend(title="Categories", loc="center left", bbox_to_anchor=(1, 0.5))

    handles = [plt.Rectangle((0, 0), 1, 1) for label in categories]
    plt.legend(
        handles,
        categories[1:11],
        bbox_to_anchor=(1.02, 0.75),
        title="Categories",
    )

    return (print(plot), print(new_table))


def img_context(df_images):

    """Function creating a plot with the number of pictures depending on their context.

    Args:
        df_images (data frame): Data frame with the information of the images/pictures used.

    Returns:
        plot (plot): Plot of the context of the images.
        table (data frame): Percentages and count of the images by context.
    """

    context_percent = round(
        (df_images["context"].value_counts() / len(df_images) * 100), 2
    )

    table = pd.DataFrame(context_percent)
    table["Count"] = df_images["context"].value_counts()
    table.rename(columns={0: "Context", "context": "Percentages"}, inplace=True)

    fig, ax = plt.subplots(figsize=(15, 7))
    dfg = df_images.groupby(["context"]).size()
    plot = dfg.plot(
        kind="bar",
        title="Size by Context",
        ylabel="Size",
        xlabel="Context",
        figsize=(6, 5),
    )

    return print(table, plot)


def img_view(df_images):

    """Function creating a plot with the number of pictures depending on their view.

    Args:
        df_images (data frame): Data frame with the information of the images/pictures used.

    Returns:
        plot (plot): Plot of the view of the images.
        table (data frame): Percentages and count of the images by view.
    """

    context_percent = round(
        (df_images["view"].value_counts() / len(df_images) * 100), 2
    )

    table = pd.DataFrame(context_percent)
    table["Count"] = df_images["view"].value_counts()
    table.rename(columns={0: "View", "view": "Percentages"}, inplace=True)

    fig, ax = plt.subplots(figsize=(15, 7))
    dfg = df_images.groupby(["view"]).size()
    plot = dfg.plot(
        kind="bar",
        title="Size by View",
        ylabel="Size",
        xlabel="View",
        figsize=(6, 5),
    )

    return (table, plot)


def img_quality(df_images):

    """Function creating a plot with the number of pictures depending on their quality.

    Args:
        df_images (data frame): Data frame with the information of the images/pictures used.

    Returns:
        plot (plot): Plot of the quality of the images.
        table (data frame): Table with the count and percentages of the images by quality.
    """

    context_percent = round(
        (df_images["image_quality"].value_counts() / len(df_images) * 100), 2
    )

    table = pd.DataFrame(context_percent)
    table["Count"] = df_images["image_quality"].value_counts()
    table.rename(columns={0: "Quality", "image_quality": "Percentages"}, inplace=True)

    fig, ax = plt.subplots(figsize=(15, 7))
    dfg = df_images.groupby(["image_quality"]).size()
    plot = dfg.plot(
        kind="bar",
        title="Size by Image Quality",
        ylabel="Size",
        xlabel="Quality",
        figsize=(6, 5),
    )

    return (table, plot)



""" The ``centernet.models`` submodule provides functions to create deep learning model for object detection.

This submodule contains the following functions:

- ``create_model(arch:str, heads:Any, head_conv:Any)`` : Create a model according the input architecture and the heads.

"""

import logging
import torch
from typing import Any
from .networks.mobilenet import get_mobilenet_v3_small

logger = logging.getLogger()


_model_factory = {
    "mobilenetv3small": get_mobilenet_v3_small,
}


def create_model(arch: str, heads: Any, head_conv: Any) -> Any:

    """Create a model according the input architecture and the heads.

    Args:
        arch (str): name of the model architecture
        heads (Any): heads of the model
        head_conv (Any): convolution head

    Returns:
        The model according the input architecture.
    """

    num_layers = int(arch[arch.find("_") + 1 :]) if "_" in arch else 0
    arch = arch[: arch.find("_")] if "_" in arch else arch
    get_model = _model_factory[arch]
    model = get_model(num_layers=num_layers, heads=heads, head_conv=head_conv)
    return model


def load_model_simple(weights="models/mobilenet_v3_pretrained.pth"):
    model = get_mobilenet_v3_small(num_layers=0, heads={"hm": 1}, head_conv=256)
    checkpoint = torch.load(weights, map_location="cpu")
    model.load_state_dict(checkpoint["model"], strict=True)
    return model



import logging
import logging.config

from flask import Flask, render_template, request

from plasticorigins.serving.config import logging_config
from plasticorigins.serving.inference import handle_post_request

logging.config.dictConfig(logging_config)

app = Flask(__name__)


@app.route("/", methods=["GET", "POST"])
def index():
    if request.method == "POST":
        return handle_post_request()
    return render_template("upload.html")


if __name__ == "__main__":
    app.run(threaded=True, port=5000, debug=False, host="0.0.0.0")



# ------------------------------------------------------------------------------
# Copyright (c) Microsoft
# Licensed under the MIT License.
# Written by Bin Xiao (Bin.Xiao@microsoft.com)
# Modified by Xingyi Zhou
# ------------------------------------------------------------------------------


import torch.nn as nn
import torchvision.models as models

BN_MOMENTUM = 0.1


class MobiletNetHM(nn.Module):
    def __init__(self, heads, head_conv, **kwargs):
        self.inplanes = 576
        self.deconv_with_bias = False
        self.heads = heads

        super().__init__()
        self.features = models.mobilenet_v3_small(pretrained=True).features

        # used for deconv layers
        self.deconv_layers = self._make_deconv_layer(
            3,
            [256, 256, 256],
            [4, 4, 4],
        )
        # self.final_layer = []

        for head in sorted(self.heads):
            num_output = self.heads[head]
            if head_conv > 0:
                fc = nn.Sequential(
                    nn.Conv2d(256, head_conv, kernel_size=3, padding=1, bias=True),
                    nn.ReLU(inplace=True),
                    nn.Conv2d(
                        head_conv,
                        num_output,
                        kernel_size=1,
                        stride=1,
                        padding=0,
                    ),
                )
            else:
                fc = nn.Conv2d(
                    in_channels=256,
                    out_channels=num_output,
                    kernel_size=1,
                    stride=1,
                    padding=0,
                )
            self.__setattr__(head, fc)

        # self.final_layer = nn.ModuleList(self.final_layer)

    def _get_deconv_cfg(self, deconv_kernel, index):
        if deconv_kernel == 4:
            padding = 1
            output_padding = 0
        elif deconv_kernel == 3:
            padding = 1
            output_padding = 1
        elif deconv_kernel == 2:
            padding = 0
            output_padding = 0

        return deconv_kernel, padding, output_padding

    def _make_deconv_layer(self, num_layers, num_filters, num_kernels):
        assert num_layers == len(
            num_filters
        ), "ERROR: num_deconv_layers is different len(num_deconv_filters)"
        assert num_layers == len(
            num_kernels
        ), "ERROR: num_deconv_layers is different len(num_deconv_filters)"

        layers = []
        for i in range(num_layers):
            kernel, padding, output_padding = self._get_deconv_cfg(num_kernels[i], i)

            planes = num_filters[i]
            layers.append(
                nn.ConvTranspose2d(
                    in_channels=self.inplanes,
                    out_channels=planes,
                    kernel_size=kernel,
                    stride=2,
                    padding=padding,
                    output_padding=output_padding,
                    bias=self.deconv_with_bias,
                )
            )
            layers.append(nn.BatchNorm2d(planes, momentum=BN_MOMENTUM))
            layers.append(nn.ReLU(inplace=True))
            self.inplanes = planes

        return nn.Sequential(*layers)

    def forward(self, x):
        x = self.features(x)
        x = self.deconv_layers(x)
        ret = {}
        for head in self.heads:
            ret[head] = self.__getattr__(head)(x)
        return ret

    def init_weights(self):
        # print('=> init resnet deconv weights from normal distribution')
        for _, m in self.deconv_layers.named_modules():
            if isinstance(m, nn.ConvTranspose2d):
                # print('=> init {}.weight as normal(0, 0.001)'.format(name))
                # print('=> init {}.bias as 0'.format(name))
                nn.init.normal_(m.weight, std=0.001)
                if self.deconv_with_bias:
                    nn.init.constant_(m.bias, 0)
            elif isinstance(m, nn.BatchNorm2d):
                # print('=> init {}.weight as 1'.format(name))
                # print('=> init {}.bias as 0'.format(name))
                nn.init.constant_(m.weight, 1)
                nn.init.constant_(m.bias, 0)
        # print('=> init final conv weights from normal distribution')
        for head in self.heads:
            final_layer = self.__getattr__(head)
            for i, m in enumerate(final_layer.modules()):
                if isinstance(m, nn.Conv2d):
                    # nn.init.kaiming_normal_(m.weight, mode='fan_out', nonlinearity='relu')
                    # print('=> init {}.weight as normal(0, 0.001)'.format(name))
                    # print('=> init {}.bias as 0'.format(name))
                    if m.weight.shape[0] == self.heads[head]:
                        if "hm" in head:
                            nn.init.constant_(m.bias, -2.19)
                        else:
                            nn.init.normal_(m.weight, std=0.001)
                            nn.init.constant_(m.bias, 0)


def get_mobilenet_v3_small(num_layers, heads, head_conv):

    model = MobiletNetHM(heads, head_conv=head_conv)
    model.init_weights()
    return model



"""The ``misc`` submodule provides computing functions for the prediction models.

This submodule contains the following class:

- ``ResizeForCenterNet`` : Resize images for Centernet model.

This submodule contains the following functions:

- ``blob_for_bbox(bbox:Union[List,array], heatmap:ndarray, downsampling_factor:Optional[float]=None)`` : Compute the 2-dimensional gaussian map (heat map) and the coordinates of the bounding box center.
- ``draw_umich_gaussian(heatmap:ndarray, center:Tuple[Union[float,int],Union[float,int]], radius:Union[float,int], k:int=1)`` : Draw the 2-dimensional gaussian map (heat map).
- ``gaussian2D(shape:Tuple[int,int], sigma:int=1)`` : Compute the 2-dimensional gaussian distribution.
- ``gaussian_radius(det_size:Tuple[int,int], min_overlap:float=0.7)`` : Compute the minimal gaussian radius.
- ``load_checkpoint(model:Any, trained_model_weights_filename:str)`` : Load the model checkpoint with specific model weights.
- ``load_model(arch:str, model_weights:Union[str,ndarray[Any,dtype[float64]]], device:str)`` : Load a prediction model with specific architecture, model weights and device type.

"""

import math
import numpy as np
from numpy import ndarray, array, dtype, float64
import torch
import torchvision.transforms.functional as F
from typing import Any, Tuple, Union, Optional, List

from plasticorigins.detection.centernet.models import create_model as create_base


class ResizeForCenterNet:

    """Resize images for Centernet model.

    Args:
        fix_res (bool): fixe resize for images with ``(H, W) = (512, 512).`` Set as default to ``False``.
    """

    def __init__(self, fix_res: bool = False):
        self.fix_res = fix_res

    def __call__(self, image: torch.Tensor) -> torch.Tensor:

        """Resize the input image according the fix_res for CenterNet model.

        Args:
            image (torch.Tensor): the input image to resize

        Returns:
            image (torch.Tensor): the resized image as output
        """

        if self.fix_res:
            new_h = 512
            new_w = 512

        else:
            w, h = image.size
            new_h = (h | 31) + 1
            new_w = (w | 31) + 1

        image = F.resize(image, (new_h, new_w))

        return image


def gaussian_radius(det_size: Tuple[int, int], min_overlap: float = 0.7) -> float:

    """Compute the minimal gaussian radius.

    Args:
        det_size (Tuple[int,int]): the size of the image / box (heigth, width)
        min_overlap (float): minimal overlap in [0,1]

    Returns:
        min_radius (float): The minimal gaussian radius
    """

    height, width = det_size

    # compute radius 1
    a1 = 1
    b1 = height + width
    c1 = width * height * (1 - min_overlap) / (1 + min_overlap)
    sq1 = np.sqrt(b1**2 - 4 * a1 * c1)
    r1 = (b1 + sq1) / 2

    # compute radius 2
    a2 = 4
    b2 = 2 * (height + width)
    c2 = (1 - min_overlap) * width * height
    sq2 = np.sqrt(b2**2 - 4 * a2 * c2)
    r2 = (b2 + sq2) / 2

    # compute radius 3
    a3 = 4 * min_overlap
    b3 = -2 * min_overlap * (height + width)
    c3 = (min_overlap - 1) * width * height
    sq3 = np.sqrt(b3**2 - 4 * a3 * c3)
    r3 = (b3 + sq3) / 2

    return min(r1, r2, r3)


def gaussian2D(shape: Tuple[int, int], sigma: int = 1) -> ndarray:

    """Compute the 2-dimensional gaussian distribution.

    Args:
        shape (Tuple[int,int]): shape of the figure
        sigma (int): standard deviation. Set as default to 1.

    Returns:
        h (ndarray): 2-dimensional gaussian array
    """

    m, n = ((ss - 1.0) / 2.0 for ss in shape)
    y, x = np.ogrid[-m : m + 1, -n : n + 1]

    h = np.exp(-(x * x + y * y) / (2 * sigma * sigma))
    h[h < np.finfo(h.dtype).eps * h.max()] = 0

    return h


def draw_umich_gaussian(
    heatmap: ndarray,
    center: Tuple[Union[float, int], Union[float, int]],
    radius: Union[float, int],
    k: int = 1,
) -> ndarray:

    """Draw the 2-dimensional gaussian heat map.

    Args:
        heatmap (ndarray): the input heat map
        center (Tuple[Union[float,int],Union[float,int]]): coordinates of the center of the 2D gaussian map
        radius (Union[float,int]): gaussian radius
        k (int): amplifier coefficient. Set as default to ``1``.

    Returns:
        masked_heatmap (ndarray) : the 2-dimensional gaussian heat map.
    """

    diameter = 2 * radius + 1
    gaussian = gaussian2D((diameter, diameter), sigma=diameter / 6)

    x, y = int(center[0]), int(center[1])

    height, width = heatmap.shape[0:2]

    left, right = min(x, radius), min(width - x, radius + 1)
    top, bottom = min(y, radius), min(height - y, radius + 1)

    masked_heatmap = heatmap[y - top : y + bottom, x - left : x + right]
    masked_gaussian = gaussian[
        radius - top : radius + bottom, radius - left : radius + right
    ]
    if min(masked_gaussian.shape) > 0 and min(masked_heatmap.shape) > 0:  # TODO debug
        np.maximum(masked_heatmap, masked_gaussian * k, out=masked_heatmap)

    return heatmap


def blob_for_bbox(
    bbox: Union[List, array],
    heatmap: ndarray,
    downsampling_factor: Optional[float] = None,
) -> Tuple[ndarray, array]:

    """Compute the 2-dimensional gaussian heat map and the coordinates of the bounding box center.

    Args:
        bbox (Union[List,array]): list of the bounding box coordinates
        heatmap (ndarray): the input heat map
        downsampling_factor (Optional[float]): downsampling factor. Set as default to ``None``.

    Returns:
        heatmap (ndarray): the 2-dimensional gaussian heat map
        ct_int (array): coordinates of the center
    """

    if downsampling_factor is not None:
        left, top, w, h = (bbox_coord // downsampling_factor for bbox_coord in bbox)

    else:
        left, top, w, h = (bbox_coord for bbox_coord in bbox)

    right, bottom = left + w, top + h
    ct_int = None

    if h > 0 and w > 0:
        radius = gaussian_radius((math.ceil(h), math.ceil(w)))
        radius = max(0, int(radius))
        ct = np.array([(left + right) / 2, (top + bottom) / 2], dtype=np.float32)
        ct_int = ct.astype(np.int32)
        heatmap = draw_umich_gaussian(heatmap, ct_int, radius)

    return heatmap, ct_int


def load_checkpoint(model: Any, trained_model_weights_filename: str) -> Any:

    """Load the model checkpoint with specific model weights.

    Args:
        model (Any): prediction model
        trained_model_weights_filename (str): file name of the model weights

    Returns:
        model (Any): prediction model from specific checkpoint
    """

    checkpoint = torch.load(trained_model_weights_filename, map_location="cpu")
    model.load_state_dict(checkpoint["model"])
    return model


def load_model(
    arch: str,
    model_weights: Union[str, ndarray[Any, dtype[float64]]],
    device: str,
) -> Any:

    """Load a prediction model with specific architecture, model weights and device type.

    Args:
        arch (str): architecture of the model ("mobilenet_v3_small", "res_18", "dla_34")
        model_weights (Union[str,ndarray[Any,dtype[float64]]]): path of the prediction model weights or model weights
        device (str): type of device used ("cpu", "cuda", ...)

    Returns:
        model (Any): prediction model with specific architecture, model weights and device type from specific checkpoint
    """

    if model_weights is None:
        if arch == "mobilenet_v3_small":
            model_weights = "models/mobilenet_v3_pretrained.pth"
            arch = "mobilenetv3small"
        elif arch == "res_18":
            model_weights = "models/res18_pretrained.pth"
        elif arch == "dla_34":
            model_weights = "models/dla_34_pretrained.pth"

    heads = {"hm": 1} if arch != "dla_34" else {"hm": 1, "wh": 2}

    model = create_base(arch, heads=heads, head_conv=256).to(device)
    model = load_checkpoint(model, model_weights)

    for param in model.parameters():
        param.requires_grad = False
    model.eval()

    return model



"""The ``detect`` submodule provides several detection functions for tracking application.

This submodule contains the following functions:

- ``detect(preprocessed_frames:Any, threshold:float, model:Any)`` : Return a detection Tensor with the localisation of the detected objects.
- ``nms(heat:torch.Tensor, kernel:int=3)`` : Applies a 2D max pooling over an input signal heat and compare the results before and after the max pooling.

"""

import torch
from numpy import ndarray
from typing import Any


def nms(heat: torch.Tensor, kernel: int = 3) -> torch.Tensor:

    """Applies a 2D max pooling over an input signal heat and compare the results before
        and after the max pooling.

    Args:
        heat (torch.Tensor): input heat map tensor (value in [0,1])
        kernel (int): kernel size for max pooling (filter kernel x kernel). Set as default to ``3`` (filter 3x3)

    Returns:
        The input heat map Tensor if the 2D max pooling is neutral. Else the null Tensor.
    """

    pad = (kernel - 1) // 2

    hmax = torch.nn.functional.max_pool2d(heat, (kernel, kernel), stride=1, padding=pad)
    keep = (hmax == heat).float()

    return heat * keep


def detect(preprocessed_frames: Any, threshold: float, model: Any) -> ndarray:

    """Return a detection Tensor with the localisation of the detected objects.

    Args:
        preprocessed_frames (Any): list of preprocessed frames from a DataLoader
        threshold (float): probability threshold for ground truths
        model (Any): Load a model pre-trained to apply to the preprocessed frames

    Returns:
        detections (ndarray): detection Tensor with the localisation of the detected objects
    """

    batch_result = torch.sigmoid(model(preprocessed_frames)["hm"])
    batch_peaks = nms(batch_result).gt(threshold).squeeze(dim=1)
    detections = [torch.nonzero(peaks).cpu().numpy()[:, ::-1] for peaks in batch_peaks]

    return detections



"""The ``inference`` submodule provides several functions for tracking and handling post request.

The submodule configures :

- the device type
- the logger
- the prediction model

This submodule contains the following functions:

- ``handle_post_request()`` : Main function to handle a post request.
- ``track(args:argparse)`` : Tracking function for object detection in frame sequences.

"""

import json
import logging
import os
from pathlib import Path
import argparse
import warnings
import numpy as np
import torch
from flask import jsonify, request
from werkzeug.utils import secure_filename
from typing import List, Tuple

# imports for tracking
from plasticorigins.detection.detect import detect
from plasticorigins.tools.files import create_unique_folder, download_from_url

from plasticorigins.tools.video_readers import IterableFrameReader
from plasticorigins.tracking.postprocess_and_count_tracks import (
    filter_tracks,
    postprocess_for_api,
)
from plasticorigins.tracking.track_video import track_video
from plasticorigins.tracking.trackers import get_tracker
from plasticorigins.tracking.utils import (
    get_detections_for_video,
    read_tracking_results,
    write_tracking_results_to_file,
)
from plasticorigins.serving.config import id_categories

# centernet / yolo version
from plasticorigins.serving.config import config_track

# from plasticorigins.serving.config import config_track_yolo as config_track

logger = logging.getLogger()
if config_track.device is None:
    device = "cuda" if torch.cuda.is_available() else "cpu"
else:
    device = config_track.device
device = torch.device(device)

engine = get_tracker("EKF")


if config_track.arch == "mobilenet_v3_small":
    from plasticorigins.tools.misc import load_model

    model = load_model(
        arch=config_track.arch,
        model_weights=config_track.model_weights,
        device=device,
    )
    logger.info("---Model mobilenet loaded.")
elif config_track.arch == "yolo":
    from plasticorigins.detection.yolo import load_model, predict_yolo

    model_path = download_from_url(
        config_track.url_model_yolo, config_track.file_model_yolo, "./models", logger
    )
    model_yolo = load_model(
        model_path,
        config_track.device,
        config_track.yolo_conf_thrld,
        config_track.yolo_iou_thrld,
    )
    logger.info("---Yolo model loaded.")
else:
    logger.error(f"unrecognized model {config_track.arch}")

observation_variance = np.load(
    os.path.join(config_track.noise_covariances_path, "observation_variance.npy")
)
transition_variance = np.load(
    os.path.join(config_track.noise_covariances_path, "transition_variance.npy")
)


def handle_post_request() -> json:

    """Main function to handle a post request. The file is in `request.files`. Will create temporary folders for storing the file and intermediate results. Outputs a json.

    Returns:
        A output json file.
    """

    logger.info("---receiving request")

    if "file" in request.files:
        file = request.files["file"]

    else:
        logger.error("error no file in request")
        return None

    # file and folder handling
    filename = secure_filename(file.filename)
    logger.info("--- received filename: " + filename)
    working_dir = Path(create_unique_folder(config_track.upload_folder, filename))
    full_filepath = working_dir / filename
    if os.path.isfile(full_filepath):
        os.remove(full_filepath)
    file.save(full_filepath)
    config_track.video_path = full_filepath.as_posix()
    config_track.output_dir = working_dir.as_posix()

    # launch the tracking
    filtered_results, num_frames, fps = track(config_track)

    # postprocess
    output_json = postprocess_for_api(filtered_results, id_categories)
    output_json["fps"] = round(fps, 2)
    output_json["video_length"] = num_frames
    output_json["video_id"] = filename

    response = jsonify(output_json)
    response.status_code = 200

    # Remove temp files (esp. the video):
    os.remove(full_filepath)

    return response


def track(args: argparse) -> Tuple[List, int, int]:

    """Tracking function for object detection in frame sequences.

    Args:
        args (argparse): arguments for tracking process

    Returns:
        filtered_results (list): list of filtered tracks
        num_frames (int): max number of frames for tracking
        fps (int): number of frames per second (video speed)
    """

    if args.arch == "mobilenet_v3_small":
        detector = lambda frame: detect(
            frame, threshold=args.detection_threshold, model=model
        )
    elif args.arch == "yolo":
        detector = lambda frame: predict_yolo(
            model_yolo, frame, size=config_track.output_shape[0], augment=False
        )
    else:
        logger.error("bad model arch")

    logger.info(f"---Processing {args.video_path}")
    reader = IterableFrameReader(
        video_filename=args.video_path,
        skip_frames=args.skip_frames,
        output_shape=args.output_shape,
        progress_bar=True,
        preload=args.preload_frames,
        crop=args.crop,
    )

    num_frames, fps = (
        int(reader.max_num_frames / (args.skip_frames + 1)),
        reader.fps,
    )

    logger.info("---Detecting...")
    detections = []
    if args.arch == "yolo":
        # Catching warnings that come from a yolo bug:
        # "User provided device_type of \'cuda\', but CUDA is not available. Disabling"
        # should be fixed with more recent versions of yolo
        with warnings.catch_warnings():
            warnings.filterwarnings("ignore")
            for frame in reader:
                detections.append(detector(frame))
    elif args.arch == "mobilenet_v3_small":
        detections = get_detections_for_video(
            reader, detector, batch_size=args.detection_batch_size, device=device
        )

    logger.info("---Tracking...")
    display = None

    results = track_video(
        reader,
        iter(detections),
        args,
        engine,
        transition_variance,
        observation_variance,
        display,
        is_yolo=args.arch == "yolo",
    )
    reader.video.release()
    # store unfiltered results
    output_filename = Path(args.output_dir) / "results_unfiltered.txt"
    coord_mapping = reader.get_inv_mapping(args.downsampling_factor)
    write_tracking_results_to_file(
        results,
        coord_mapping,  # Scale the output back to original video size
        output_filename=output_filename,
    )
    logger.info("---Filtering...")

    # read from the file
    results = read_tracking_results(output_filename)
    filtered_results = filter_tracks(results, args.kappa, args.tau)
    # store filtered results
    output_filename = Path(args.output_dir) / "results.txt"
    write_tracking_results_to_file(
        filtered_results,
        lambda x, y: (x, y),  # No scaling, already scaled!
        output_filename=output_filename,
    )

    return filtered_results, num_frames, fps



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['plasticorigins',
 'plasticorigins.detection',
 'plasticorigins.detection.centernet',
 'plasticorigins.detection.centernet.networks',
 'plasticorigins.serving',
 'plasticorigins.tools',
 'plasticorigins.tracking',
 'plasticorigins.training',
 'plasticorigins.training.data',
 'plasticorigins.training.docs',
 'plasticorigins.training.visualization']

package_data = \
{'': ['*'], 'plasticorigins.serving': ['templates/*']}

install_requires = \
['Flask==2.0.3',
 'Pillow==9.1.1',
 'Werkzeug==2.0.3',
 'debugpy==1.5.1',
 'gunicorn==19.10.0',
 'imgaug==0.4.0',
 'opencv-python==4.5.5.62',
 'pandas==1.5.0',
 'psycopg2-binary==2.9.3',
 'pycocotools==2.0.4',
 'pykalman>=0.9.5,<0.10.0',
 'scikit-learn>=1.1.0,<2.0.0',
 'scikit-video>=1.1.11,<2.0.0',
 'scipy==1.7.3',
 'torch>=1.11.0,<2.0.0',
 'torchvision>=0.12.0,<0.13.0',
 'tqdm==4.62.3',
 'yolov5>=6.1.6,<7.0.0']

setup_kwargs = {
    'name': 'plastic-origins',
    'version': '2.0.0',
    'description': 'A package containing methods commonly used to make inferences',
    'long_description': "# Automated object counting on riverbanks\n\n## Project Branches:\n\n### release:\n\nThis is the main/production branch (DO NOT PUSH DIRECTLY TO RELEASE)\n### dev: \n\nThis is the developpement branch (DO NOT PUSH DIRECTLY TO DEV)\n### Feature Branches\n\nThese are the branches where you can develop new features for the project. In order to create a feature branch:\n\n- Make sure that your **dev** branch is up to date\n    ```shell\n    git checkout dev\n    git pull dev\n    ```\n- Create a new branch from **dev** with the name **feature/name_of_your_feature**\n    ```shell\n    git checkout -b feature/name_of_your_feature\n    ```\n- Once your feature developpement is complete, make a Pull Request of your feature branch to **dev**\n### Research Branches: \n\nThese are branches made for research purposes and they are named **research/name_of_your_subject**\n## Release Branch - Installation\n\nFollow these steps in that order exactly:\n\n### Clone the project\n```shell\ngit clone https://github.com/surfriderfoundationeurope/surfnet.git <folder-for-surfnet> -b release\ncd <folder-for-surfnet>\n```\n### Install Poetry\n```shell\npip install poetry\n```\n\n### Create your virtual environment\nHere we use python version 3.9\n```shell\npoetry use 3.9\n```\n\n### Install dependencies\n```shell\npoetry install\n```\n\n### Code Linting and Formatting:\n\npre-commits have been added to format and check the linting of the code before any commit. This process will run:\n- PyUpgrade: to make sure that the code syntax is up to date with the latest python versions\n- Black: which is a code formatter \n- Flake8: to check that the code is properly formatted.\n\nAll this process is automatic to ensure the commited code quality. So as a good measure, prior to committing any code it is highly recommended to run:\n```shell\npoetry run black path/to/the/changed/code/directory(ies)\n```\nThis will format the code that has been written and:\n```shell\npoetry run flake8 path/to/the/changed/code/directory(ies)\n```\nto check if there is any other issues to fix.\n## Downloading pretrained models\n\nYou can download MobileNetV3 model with the following script:\n```shell\ncd models\nsh download_pretrained_base.sh\n```\nThe file will be downloaded into [models](models).\n\n## Validation videos\n\nIf you want to download the 3 test videos on the 3 portions of the Auterrive riverbank, run:\n\n```\ncd data\nsh download_validation_videos.sh\n```\n\nThis will download the 3 videos in distinct folders of [data/validation_videos](data/validation_videos).\n\n## Serving\n\n### Development\nSetting up the server and testing: from surfnet/ directory, you may run a local flask developement server with the following command:\n\n```shell\nexport FLASK_APP=src/plasticorigins/serving/app.py\npoetry run flask run\n```\n\n### Production\nSetting up the server and testing: from surfnet/ directory, you may run a local wsgi gunicorn production server with the following command:\n\n```shell\nPYTHONPATH=./src gunicorn -w 5 --threads 2 --bind 0.0.0.0:8001 --chdir ./src/serving/ wsgi:app\n```\n\n### Test surfnet API\nThen, in order to test your local dev server, you may run:\n```shell\ncurl -X POST http://127.0.0.1:5000/ -F 'file=@/path/to/video.mp4' # flask\n```\nChange port 5000 to 8001 to test on gunicorn or 8000 to test with Docker and gunicorn.\n\n### Docker\nYou can build and run the surfnet AI API within a Docker container.\n\nDocker Build:\n```shell\ndocker build -t surfnet/surfnet:latest .\n```\n\nDocker Run:\n```shell\ndocker run --env PYTHONPATH=/src -p 8000:8000 --name surfnetapi surfnet/surfnet:latest\n```\n\n### Makefile\nYou can use the makefile for convenience purpose to launch the surfnet API:\n```shell\nmake surfnet-dev-local # with flask\nmake surfnet-prod-local # with gunicorn\nmake surfnet-prod-build-docker # docker build\nmake surfnet-prod-run-docker # docker run\n```\n\n### Kubernetes\nTo ease production operation, the surfnet API can be deployed on top of kubernetes (k8s) cluster. A pre-built Docker image is available on ghcr.io to be deployed using the surfnet.yaml k8s deployment file. To do so, change directory to k8s/, then once you are connected to your k8s cluster simply enter:\n```shell\nkubectl apply -y surfnet.yaml\n```\nRemark: we use a specific surfnet k8s node pool label for our Azure production environment on aks. If you want to test deployment on a default k8s cluster using system nodes, you have either to use default surfnet.yaml file or remove the nodeSelector section from others deployment files (aks, gke).\n\nAfter the deployment is done, create a service to expose the surfnet API to be publicly accessible over the Internet.\n```shell\nkubectl expose deployment surfnet --type=LoadBalancer --name=surfnet-api\nkubectl get service surfnet-api\n```\n\n## Release plasticorigins to pypi:\n\n### Prerelease: (Pull Request to Dev branch)\n#### Check or Bump version:\n\nCheck the current version of the product:\n\nDocker Build:\n```shell\npoetry version\n```\n\nBump the version to the product:\n\n```shell\npoetry version <bump-rule>\n```\nbump rules can be found in : https://python-poetry.org/docs/cli/#:~:text=with%20concrete%20examples.-,RULE,-BEFORE\n**choose carefully the one that corresponds to your bump: for the prerelease we will use:**\n- **prepatch**\n- **preminor**\n- **premajor**\n\nmake sure that in you pyproject.toml your version ends with **-alpha.0**\n\n#### Publish the prerelease to pypi:\n\nIn order to publish your prerelease to PyPi, all you need to do is open a Pull Request of your current branch to **Dev** branch. Once the PR is approved and merged, the Prerelease will be done automatically with a github workflow.\n\n### Release: (Pull Request to Release branch):\nIn order to publish a release version to PyPi, all you have to do is open a Pull Request of the **Dev** branch into the **Release** branch. Once the PR is approved and merged, the Release will be done automatically with a github workflow.\n\n## Testing:\nTo launch the tests you can run this command\n```shell\npoetry run coverage run -m pytest -s && poetry run coverage report -m\n```\n\n## Mkdocs Documentation:\nYou need to install the following packages:\n```shell\npip install mkdocs\npip install mkdocstrings\n```\nTo run the mkdocs documentation, you can run the following lines below:\n```shell\ncd src\nmkdocs serve\n```\nThe documentation will be serving on http://127.0.0.1:8000/.\n\n## Configuration\n\n`src/serving/inference.py` contains a Configuration dictionary that you may change:\n- `skip_frames` : `3` number of frames to skip. Increase to make the process faster and less accurate.\n- `kappa`: `7` the moving average window. `1` prevents the average, avoid `2` which is ill-defined.\n- `tau`: `4` the number of consecutive observations necessary to keep a track. If you increase `skip_frames`, you should lower `tau`.\n\n## Datasets and Training\n\nConsider other branches for that!\n",
    'author': 'Chayma Mesbahi',
    'author_email': 'chayma.mesbahi@neoxia.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://www.plasticorigins.eu/',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<3.11',
}


setup(**setup_kwargs)



"""The ``track_video`` submodule provides several functions to track trashs on videos.

This submodule contains the following functions :

- associate_detections_to_trackers(detections_for_frame:List[ndarray], confs:array[Any,dtype[float64]], labels:array[Any,dtype[int64]],
    trackers:Dict[int,Tracker], flow01:ndarray, confidence_threshold:float) : Associate detections to trackers.
- build_confidence_function_for_trackers(trackers:Dict[int,Tracker], flow01:ndarray) : Build confidence function for trackers.
- init_trackers(engine:Any, detections:List[ndarray], confs:array[Any,dtype[float64]], labels:array[Any,dtype[int64]], frame_nb:int, state_variance:ndarray[Any,dtype[float64]], observation_variance:ndarray[Any,dtype[float64]],
    delta:float): Initializes the trackers based on detections.
- interpret_detection(detections_for_frame:List[ndarray], downsampling_factor:Union[int,float], is_yolo:bool=False) : Normalizes the detections depending whether they come from Centernet or Yolo model
- track_video(reader:Any, detections:List[ndarray], args:argparse, engine:Any, transition_variance:ndarray[Any,dtype[float64]],
    observation_variance:ndarray[Any,dtype[float64]], display:Any, is_yolo:bool=False) : Original version of tracking trashs on video.

"""

import argparse
import numpy as np
from numpy import array, ndarray, dtype, float64
from typing import Any, Dict, Iterable, List, Tuple, Union

from scipy.optimize import linear_sum_assignment
from scipy.spatial.distance import euclidean

from plasticorigins.tools.optical_flow import compute_flow
from plasticorigins.tracking.utils import in_frame
from plasticorigins.tracking.trackers import Tracker


def init_trackers(
    engine: Any,
    detections: List[ndarray],
    confs: array,
    labels: array,
    frame_nb: int,
    state_variance: ndarray[Any, dtype[float64]],
    observation_variance: ndarray[Any, dtype[float64]],
    delta: float,
) -> List[Tracker]:

    """Initializes the trackers based on detections.

    Args:
        engine (Any): the engine tracker
        detections (List[ndarray]): list of detected object positions with center coordinates such as ``list[np.array([[xcenter, ycenter], ...]), ...]``
        confs (array[Any,dtype[float64]]): array of the confidence values
        labels (array[Any,dtype[int64]]): array of the object classes
        frame_nb (int): number of frames
        state_variance (ndarray[Any,dtype[float64]]): state variances
        observation_variance (ndarray[Any,dtype[float64]]): observation variances
        delta (float): threshold in probabilty between [0,1] (confidence threshold). Probabilty of a point to belong to a specific neighbourhood.

    Returns:
        trackers (List[Tracker]): the trackers based on detections and previous parameters
    """

    trackers = []

    for detection, conf, label in zip(detections, confs, labels):
        tracker_for_detection = engine(
            frame_nb,
            detection,
            conf,
            label,
            state_variance,
            observation_variance,
            delta,
        )
        trackers.append(tracker_for_detection)

    return trackers


def build_confidence_function_for_trackers(
    trackers: Dict[int, Tracker], flow01: ndarray
) -> Tuple[List[int], List]:

    """Build confidence function for trackers.

    Args:
        trackers (Dict[int,Tracker]): the trackers
        flow01 (ndarray): the reference flow

    Returns:
        tracker_nbs (List[int]): the list of tracker numbers
        confidence_functions (List[Any]): the list of confidence functions
    """

    tracker_nbs = []
    confidence_functions = []

    for tracker_nb, tracker in enumerate(trackers):
        if tracker.enabled:
            tracker_nbs.append(tracker_nb)
            confidence_functions.append(tracker.build_confidence_function(flow01))

    return tracker_nbs, confidence_functions


def associate_detections_to_trackers(
    detections_for_frame: List[ndarray],
    confs: array,
    labels: array,
    trackers: Dict[int, Tracker],
    flow01: ndarray,
    confidence_threshold: float,
) -> List:

    """Associate detections to trackers.

    Args:
        detections_for_frame (List[np.ndarray]): list of detected object positions with center coordinates such as ``list[np.array([[xcenter, ycenter], ...]), ...]``
        confs (array[float]): array of the confidence values
        labels (array[int]): array of the object labels
        trackers (Dict[int,Tracker]): the trackers
        flow01 (array): the reference flow
        confidence_threshold (float): confidence threshold for detection in [0,1]

    Returns:
        assigned_trackers (List[int]): the list of assigned trackers
    """

    tracker_nbs, confidence_functions = build_confidence_function_for_trackers(
        trackers, flow01
    )
    assigned_trackers = [None] * len(detections_for_frame)

    if len(tracker_nbs):
        cost_matrix = np.zeros(shape=(len(detections_for_frame), len(tracker_nbs)))

        for detection_nb, (detection, conf, label) in enumerate(
            zip(detections_for_frame, confs, labels)
        ):

            for tracker_id, confidence_function in enumerate(confidence_functions):
                score = confidence_function(detection)
                cls_score = trackers[tracker_id].cls_score_function(conf, label)

                if cls_score < 0.5:
                    score = score * 0.1  # if wrong class, reduce the score, to tweak

                if score > confidence_threshold:
                    cost_matrix[detection_nb, tracker_id] = score

                else:
                    cost_matrix[detection_nb, tracker_id] = 0

        row_inds, col_inds = linear_sum_assignment(cost_matrix, maximize=True)

        for row_ind, col_ind in zip(row_inds, col_inds):
            if cost_matrix[row_ind, col_ind] > confidence_threshold:
                assigned_trackers[row_ind] = tracker_nbs[col_ind]

    return assigned_trackers


def interpret_detection(
    detections_for_frame: List[ndarray],
    downsampling_factor: Union[int, float],
    is_yolo: bool = False,
) -> Tuple[List[ndarray], array, array]:

    """Normalizes the detections depending whether they come from Centernet or Yolo model

    Args:
        detections_for_frame (List[ndarray]): list of detected object positions with center coordinates such as ``list[np.array([[xcenter, ycenter], ...]), ...]``
        downsampling_factor (Union[int,float]): downsampling factor
        is_yolo (bool): ``True`` if the model used for detection is Yolo. ``False`` if the model used for detection is Centernet.

    Returns:
        detections_for_frame (List[ndarray]): list of detected object positions with center coordinates such as ``list[np.array([[xcenter, ycenter], ...]), ...]``
        confs (array[Any,dtype[float64]]): array of the confidence values after normalization
        labels (array[Any,dtype[int64]]): array of the object labels after normalization
    """

    if not is_yolo:
        confs = [1.0] * len(detections_for_frame)
        labels = [0] * len(detections_for_frame)
        return detections_for_frame, confs, labels

    else:
        detections_for_frame, confs, labels = detections_for_frame
        # get center
        detections_for_frame = detections_for_frame[..., 0:2] / downsampling_factor
        return detections_for_frame, confs, labels


def track_video(
    reader: Any,
    detections: List[ndarray],
    args: argparse,
    engine: Any,
    transition_variance: ndarray[Any, dtype[float64]],
    observation_variance: ndarray[Any, dtype[float64]],
    display: Any,
    is_yolo: bool = False,
) -> Iterable:

    """Original version of tracking trashs on video. Expects detections in the format ``list[np.array([[xcenter, ycenter], ...]), ...]``.

    Args:
        reader (Any): the video reader
        detections (List[ndarray]): array of detected object positions
        args (argparse): array of the confidence values
        engine (Any): the engine tracker
        transition_variance (ndarray[Any,dtype[float64]]): transition variances
        observation_variance (ndarray[Any,dtype[float64]]): observation variances
        display (Any): to display the tracker
        is_yolo (bool): ``True`` if the model used for detection is Yolo. ``False`` if the model used for detection is Centernet.

    Returns:
        results (Iterable): the tracker based on detections and previous parameters
    """

    init = False
    trackers = dict()
    frame_nb = 0
    frame0 = next(reader)
    detections_for_frame = next(detections)
    detections_for_frame, confs, labels = interpret_detection(
        detections_for_frame, args.downsampling_factor, is_yolo
    )

    max_distance = euclidean(reader.output_shape, np.array([0, 0]))
    delta = 0.005 * max_distance

    if display is not None and display.on:
        display.display_shape = (
            reader.output_shape[0] // args.downsampling_factor,
            reader.output_shape[1] // args.downsampling_factor,
        )
        display.update_detections_and_frame(detections_for_frame, frame0)

    if len(detections_for_frame):
        trackers = init_trackers(
            engine,
            detections_for_frame,
            confs,
            labels,
            frame_nb,
            transition_variance,
            observation_variance,
            delta,
        )
        init = True

    if display is not None and display.on:
        display.display(trackers)

    for frame_nb, (frame1, detections_for_frame) in enumerate(
        zip(reader, detections), start=1
    ):
        detections_for_frame, confs, labels = interpret_detection(
            detections_for_frame, args.downsampling_factor, is_yolo
        )

        if display is not None and display.on:
            display.update_detections_and_frame(detections_for_frame, frame1)

        if not init:
            if len(detections_for_frame):
                trackers = init_trackers(
                    engine,
                    detections_for_frame,
                    confs,
                    labels,
                    frame_nb,
                    transition_variance,
                    observation_variance,
                    delta,
                )
                init = True
        else:
            new_trackers = []
            flow01 = compute_flow(frame0, frame1, args.downsampling_factor)

            if len(detections_for_frame):
                assigned_trackers = associate_detections_to_trackers(
                    detections_for_frame,
                    confs,
                    labels,
                    trackers,
                    flow01,
                    args.confidence_threshold,
                )

                for detection, conf, label, assigned_tracker in zip(
                    detections_for_frame, confs, labels, assigned_trackers
                ):
                    if in_frame(detection, flow01.shape[:-1]):
                        if assigned_tracker is None:
                            new_trackers.append(
                                engine(
                                    frame_nb,
                                    detection,
                                    conf,
                                    label,
                                    transition_variance,
                                    observation_variance,
                                    delta,
                                )
                            )
                        else:
                            trackers[assigned_tracker].update(
                                detection, conf, label, flow01, frame_nb
                            )

            for tracker in trackers:
                tracker.update_status(flow01)

            if len(new_trackers):
                trackers.extend(new_trackers)

        if display is not None and display.on:
            display.display(trackers)
        frame0 = frame1.copy()

    results = []
    tracklets = [tracker.tracklet for tracker in trackers]

    for tracker_nb, dets in enumerate(tracklets):
        for det in dets:
            results.append((det[0], tracker_nb, det[1][0], det[1][1], det[2], det[3]))

    results = sorted(results, key=lambda x: x[0])

    return results
